#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 13 15:11:12 2020

@author: psakic

This sub-module of geodezyx.conv deals with angle conversion.

it can be imported directly with:
from geodezyx import conv

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

#### Import the logger
import logging
import re

########## BEGIN IMPORT ##########
#### External modules
import numpy as np

#### geodeZYX modules
from geodezyx import utils

log = logging.getLogger("geodezyx")

##########  END IMPORT  ##########

#                       _         _____                              _
#     /\               | |       / ____|                            (_)
#    /  \   _ __   __ _| | ___  | |     ___  _ ____   _____ _ __ ___ _  ___  _ __
#   / /\ \ | '_ \ / _` | |/ _ \ | |    / _ \| '_ \ \ / / _ \ '__/ __| |/ _ \| '_ \
#  / ____ \| | | | (_| | |  __/ | |___| (_) | | | \ v /  __/ |  \__ \ | (_) | | | |
# /_/    \_\_| |_|\__, |_|\___|  \_____\___/|_| |_|\_/ \___|_|  |___/_|\___/|_| |_|
#                  __/ |
#                 |___/


### Angle conversion
def degdec2dms(deg_in, only_dm=False):
    """
    Angle representation conversion

    Convert :
    Decimal Degree => Degree Minute Seconds

    See also dms_num2str for Numeric => String representation conversion

    Parameters
    ----------
    deg_in : float or iterable of floats
        Decimal degrees

    only_dm : bool
        True to get the output string as a Degree Minute Angle only

    Returns
    -------
    deg , minu , sec : np.array of float
        3 arrays for Degrees, Minutes, Seconds

    """
    # bring everything to positive
    sign = np.sign(deg_in)
    deg_pos = np.abs(deg_in)

    deg = np.floor(deg_pos)
    decimal_part = deg_pos - deg
    decimal_part_sec = decimal_part * 3600
    minu = np.floor_divide(decimal_part_sec, 60)
    sec = decimal_part_sec - minu * 60
    sec = np.round(sec, 10)  # ROUNDING is it necessary?

    # back to the right hemisphere
    deg = sign * deg

    if only_dm:
        return deg, minu + sec * (1.0 / 60.0)
    else:
        return deg, minu, sec


def dms_num2str(dms_inp, decimal_fmt="6.3f"):
    """
    Angle representation conversion

    Convert :
    numeric Degree-Minute-Second (DMS) angle => a string representation.

    Parameters
    ----------
    dms_inp : list or tuple of floats
        A list or tuple containing the DMS values. The first element is degrees,
        the second element (optional) is minutes, and the third element (optional) is seconds.
    decimal_fmt : str, optional
        The format for the seconds part of the DMS string. Default is '7.4f'.

    Returns
    -------
    out_str : str
        A string representation of the DMS angle in the format "DD°MM'SS.SSSS\"".
        If only degrees and minutes are provided, the format will be "DD°MM'".
    """

    out_str = "{:02d}".format(int(dms_inp[0])) + "°"

    decimal_fmt_use = "{:0" + decimal_fmt + "}"

    if len(dms_inp) == 3:
        out_str = (
            out_str
            + "{:02d}".format(int(dms_inp[1]))
            + "'"
            + decimal_fmt_use.format(dms_inp[2])
            + '"'
        )
    elif len(dms_inp) == 2:
        out_str = out_str + decimal_fmt_use.format(dms_inp[1]) + "'"

    return out_str


def dms2degdec_num(deg, minn=0, sec=0):
    """
    Angle representation conversion

    Convert :
    Degree Minute Second `float` Angle => decimal Degree `float` Angle

    Parameters
    ----------
    deg & minn & sec : float
        degres, minutes and seconds of the input Angle

    Returns
    -------
    dd_float : float
        Decimal degree Angle

    """
    sig = np.sign(deg)

    return deg + sig * minn * (1.0 / 60.0) + sig * sec * (1.0 / 3600.0)


def dms2degdec_str(dms_str, only_dm=False):
    """
    Angle representation conversion

    Convert :
    Degree Minute Second `string` Angle => decimal Degree `float` Angle

    can manage only DM angle in input

    NB : Too Complicated .... must be simplified

    Parameters
    ----------
    dms_str : str
        string of DMS angle
        e.g.
        "2°20'35.09"E"

    only_dm : bool
        True if the string is only a Degree Minute Angle
        e.g.
        "40°52.0931'N"
        "28°31.4136'E"

    Returns
    -------
    dd_float : float
        Decimal degree Angle

    """

    if not only_dm:
        log.warning("DMS mode not well implemented yet !!! ")

    dms_str = dms_str.strip()

    if re.match(".*[swoSWO].*", dms_str):
        sign = -1
    else:
        sign = 1

    # former regex
    lis = re.findall(r"\d+|\D+", dms_str)

    ipt = lis.index(".")
    decimal = lis[ipt - 1] + lis[ipt] + lis[ipt + 1]
    lis[ipt - 1] = decimal
    lis = lis[:ipt]

    for i, e in enumerate(lis):
        try:
            lis[i] = float(e)
        except:
            lis.remove(e)

    lis = [float(e) for e in lis]

    deg = lis[0]
    minu = lis[1]
    if only_dm:
        sec = 0.0
    else:
        try:
            sec = lis[2]
        except IndexError:
            log.error("did you forgot to activate the DM only mode ? ;) ?")
            return None

    dd_float = sign * (deg + minu / 60.0 + sec / 3600.0)
    return dd_float


###### arcsec


def arcsec2deg(arcsec_in):
    """
    Angle conversion

    Convert :
    Arcsecond => Degrees

    NB : Not really useful, should be avoided

    """

    return arcsec_in / 3600.0


def deg2arcsec(deg_in):
    """
    Angle conversion

    Convert :
    Degrees => Arcsecond

    NB : Not really useful, should be avoided

    """
    return deg_in * 3600.0


def angle2equivalent_earth_radius(angle_in, angtype="deg", earth_radius=6371008.8):
    """
    Quick and simple function which gives the equivalent distance on a
    spherical Earth great circle of an angle

    Useful to determine metric varations in latitude

    angle can be : "deg", "rad", "mas"

    """
    earth_circum = earth_radius * 2 * np.pi
    if angtype == "deg":
        equiv_out = (angle_in * earth_circum) / 360.0
    elif angtype == "rad":
        equiv_out = (angle_in * earth_circum) / (np.pi * 2)
    elif angtype == "mas":
        equiv_out = (angle_in * 10**-3 * earth_circum) / 86400.0
    else:
        log.error("bad angtype %s", angtype)
        return None

    return equiv_out


def angle2equivalent_earth_parallel(
    angle_in, latitude_in, angtype="deg", earth_radius=6371008.8
):
    """
    Quick and simple function which gives the equivalent distance on an
    Earth parallel circle of an angle

    Useful to determine metric varaitions in longitude
    (but the latitude is also necessary to determine
     the radius of the parallel)

    angle can be : "deg", "rad", "mas"

    """

    if angtype == "deg":
        latitude_rad = np.deg2rad(latitude_in)
    elif angtype == "rad":
        latitude_rad = latitude_in
    elif angtype == "mas":
        log.warning("double check the mas conversion for security")
        latitude_rad = (
            2 * np.pi * (86400.0 * 10**3)
        ) / latitude_in  ### a verifier ....
    else:
        log.error("bad angtype %s", angtype)
        return None

    parallel_radius = np.cos(latitude_rad) * earth_radius

    return angle2equivalent_earth_radius(angle_in, angtype, parallel_radius)


def anglesfromvects(xa, ya, xb, yb, angtype="deg"):
    """
    Determines the angle between the points A(xa,ya) and B(xb,yb)

    angle can be : "deg", "rad"

    """
    a = np.array([xa, ya])
    b = np.array([xb, yb])

    ps = np.inner(a, b)
    a = np.arccos(ps / (np.linalg.norm(a) * np.linalg.norm(b)))

    if angtype == "deg":
        return np.rad2deg(a)
    elif angtype == "rad":
        return a
    else:
        log.error("bad angtype")
        return None


def angle_interpolation_quick(a, b, w):
    """
    Determine the interpolation between angle A & B
    by conversion to the cartesian space
    the parameter w € [0,1] define the interpoled angle C(w)
    where c(w=0) = a  &  c(w=1) = b

    References
    ----------
    https://stackoverflow.com/questions/2708476/rotation-interpolation

    """

    cs = (1 - w) * np.cos(a) + w * np.cos(b)
    sn = (1 - w) * np.sin(a) + w * np.sin(b)
    c = np.arctan2(sn, cs)

    return c


def angle_from_3_pts(p1, p2, p3):
    """
    Gives angle between 3 points
    p3 is the vertex (sommet)

    References
    ----------
    http://www.les-mathematiques.net/phorum/read.php?8,596072,596231

    """

    p1, p2, p3 = [np.array(e) for e in (p1, p2, p3)]
    x1, y1 = p1
    x2, y2 = p2
    x3, y3 = p3

    kos = ((x1 - x3) * (x2 - x3) + (y1 - y3) * (y2 - y3)) / (
        np.linalg.norm(p1 - p3) * np.linalg.norm(p2 - p3)
    )

    return np.arccos(kos)


def cartesian2polar(x, y):
    """
    Coordinates conversion

    cartesian => polar conversion

    Parameters
    ----------
    x , y : numpy.array of float
        cartesian coordinates

    Returns
    -------
    r , theta : float
        polar coordinates (radius / angle in radians)
    """
    if utils.is_iterable(x):
        x = np.array(x)
    if utils.is_iterable(y):
        y = np.array(y)

    theta = np.arctan2(y, x)
    r = np.sqrt(x ** +(y**2))
    return r, theta


def polar2cartesian(r, theta, ang="deg"):
    """
    Coordinates conversion

    polar => cartesian conversion

    Parameters
    ----------
    r , theta : float or iterable of floats
        polar coordinates

    ang : string
        'deg' (degrees) or 'rad' (radian)

    Returns
    -------
    x , y : numpy.array of float
        cartesian coordinates
    """
    if utils.is_iterable(r):
        r = np.array(r)
    if utils.is_iterable(theta):
        theta = np.array(theta)

    if ang == "deg":
        theta = np.deg2rad(theta)
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return x, y



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.files_rw contains functions to
write misc. geodetic data in dedicated files.

it can be imported directly with:
from geodezyx import files_rw

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt

#### Import the logger
import logging
import os
import re

import numpy as np
import pandas as pd

#### geodeZYX modules
from geodezyx import conv
from geodezyx import files_rw
from geodezyx import reffram
from geodezyx import utils
from geodezyx import operational

# from geodezyx.megalib.megalib import *   # Import the legacy modules names
log = logging.getLogger("geodezyx")


#### Import star style
# from geodezyx import *                   # Import the GeodeZYX modules
# from geodezyx.externlib import *         # Import the external modules
# from geodezyx.megalib.megalib import *   # Import the legacy modules names
##########  END IMPORT  ##########


def write_sp3(
    orb_df_in,
    outpath,
    outname=None,
    prefix="orb",
    skip_null_epoch=True,
    force_format_c=False,
):
    """
    Write a SP3 file from an Orbit DataFrame

    Parameters
    ----------
    orb_df_in : DataFrame
        Input Orbit DataFrame.
    outpath : str
        The output path of the file (see also outname).
    outname : None or str, optional
        None = outpath is the full path (directory + filename) of the output.
        A string = a manual name for the file.
        'auto_old_cnv' = automatically generate the filename (old convention)
        'auto_new_cnv' = automatically generate the filename (new convention)
        The default is None.
    prefix : str, optional
        the output 3-char. name of the AC. The default is 'orb'.
    skip_null_epoch : bool, optional
        Do not write an epoch if all sats are null (filtering).
        The default is True.
    force_format_c : bool, optional
        force SP3's format c. The default is False.

    Returns
    -------
    The string containing the formatted SP3 data.
    """

    ################## MAIN DATA
    lines_stk = []

    sp3_df_wrk = orb_df_in.sort_values(["epoch", "prn"])

    epoch_raw_list = sp3_df_wrk["epoch"].unique()
    sat_list = sorted(sp3_df_wrk["prn"].unique())
    sat_list = list(sorted(sp3_df_wrk["prn"].unique()))
    ## sat_list    = list(reversed(sat_list))
    #### PS 210721
    #### reversed bc the sats are sorted ascending=False, but why???
    #### to have G before E ??
    sat_list_set = set(sat_list)
    epoch_used_list = []

    if not "clk" in sp3_df_wrk.columns:
        sp3_df_wrk["clk"] = 999999.999999

    for epoc in epoch_raw_list:
        sp3epoc = pd.DataFrame(sp3_df_wrk[sp3_df_wrk["epoch"] == epoc])

        ######## if keep_missing_sat_in_epoch:
        ## manage missing sats for the current epoc
        missing_sats = sat_list_set.difference(set(sp3epoc["prn"]))

        for miss_sat in missing_sats:
            miss_line = sp3epoc.iloc[0].copy()
            miss_line["prn"] = miss_sat
            miss_line["sys"] = miss_sat[0]
            ### check the sp3 doc
            # bad position = 0.000000
            # bad clock    = 999999.9999999
            miss_line["x"] = 0.000000
            miss_line["y"] = 0.000000
            miss_line["z"] = 0.000000
            miss_line["clk"] = 999999.999999

            sp3epoc = sp3epoc.append(miss_line)
        #### end of missing sat bloc

        sp3epoc.sort_values("prn", inplace=True, ascending=True)
        timestamp = conv.dt2sp3_timestamp(conv.numpy_dt2dt(epoc)) + "\n"

        linefmt = "p{:}{:14.6f}{:14.6f}{:14.6f}{:14.6f}\n"

        LinesStkEpoch = []
        sum_val_epoch = 0
        for ilin, lin in sp3epoc.iterrows():
            if not "clk" in lin.index:  # manage case if no clk in columns
                lin["clk"] = 999999.999999
            line_out = linefmt.format(
                lin["prn"], lin["x"], lin["y"], lin["z"], lin["clk"]
            )

            sum_val_epoch += lin["x"] + lin["y"] + lin["z"]

            LinesStkEpoch.append(line_out)

        ### if skip_null_epoch activated, print only if valid epoch
        if not (np.isclose(sum_val_epoch, 0) and skip_null_epoch):
            lines_stk.append(timestamp)  # stack the timestamp
            lines_stk = lines_stk + LinesStkEpoch  # stack the values
            epoch_used_list.append(epoc)  # stack the epoc as dt

    ################## HEADER
    ######### SATELLITE LIST

    satline_stk = []
    sigmaline_stk = []

    if force_format_c:
        nlines = 5
    else:
        div, mod = np.divmod(len(sat_list), 17)

        if div < 5:
            nlines = 5
        else:
            nlines = div

            if mod != 0:
                nlines += 1

    for i in range(nlines):
        sat_line = sat_list[17 * i : 17 * (i + 1)]
        sat_line_sigma = len(sat_line) * " 01"

        if len(sat_line) < 17:
            complem = " 00" * (17 - len(sat_line))
        else:
            complem = ""

        if i == 0:
            nbsat4line = len(sat_list)
        else:
            nbsat4line = ""

        satline = "+  {:3}   ".format(nbsat4line) + "".join(sat_line) + complem + "\n"
        sigmaline = "++         0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0\n"
        sigmaline = "++       " + sat_line_sigma + complem + "\n"

        satline_stk.append(satline)
        sigmaline_stk.append(sigmaline)

    ######### 2 First LINES
    start_dt = conv.numpy_dt2dt(np.min(epoch_used_list))

    header_line1 = (
        "#cP"
        + conv.dt2sp3_timestamp(start_dt, False)
        + "     {:3}".format(len(epoch_used_list))
        + "   u+U IGSXX FIT  XXX\n"
    )

    delta_epoch = int(utils.most_common(np.diff(epoch_used_list) * 10**-9))
    mjd = conv.dt2mjd(start_dt)
    mjd_int = int(np.floor(mjd))
    mjd_dec = mjd - mjd_int
    gps_wwww, gps_sec = conv.dt2gpstime(start_dt, True, "gps")

    header_line2 = "## {:4} {:15.8f} {:14.8f} {:5} {:15.13f}\n".format(
        gps_wwww, gps_sec, delta_epoch, mjd_int, mjd_dec
    )

    ######### HEADER BOTTOM
    header_bottom = """%c M  cc GPS ccc cccc cccc cccc cccc ccccc ccccc ccccc ccccc
%c cc cc ccc ccc cccc cccc cccc cccc ccccc ccccc ccccc ccccc
%f  1.2500000  1.025000000  0.00000000000  0.000000000000000
%f  0.0000000  0.000000000  0.00000000000  0.000000000000000
%i    0    0    0    0      0      0      0      0         0
%i    0    0    0    0      0      0      0      0         0
/* PCV:IGSXX_XXXX OL/AL:FESXXXX  NONE     YN CLK:CoN ORB:CoN
/*     GeodeZYX Toolbox Output
/*
/*
"""

    ################## FINAL STACK

    final_lines_stk = []

    final_lines_stk.append(header_line1)
    final_lines_stk.append(header_line2)
    final_lines_stk = final_lines_stk + satline_stk + sigmaline_stk
    final_lines_stk.append(header_bottom)
    final_lines_stk = final_lines_stk + lines_stk + ["EOF"]

    final_str = "".join(final_lines_stk)

    ### Manage the file path
    prefix_opera = prefix

    if not outname:
        outpath_opera = outpath
    elif outname == "auto_old_cnv":
        week, dow = conv.dt2gpstime(start_dt)
        filename = prefix_opera + str(week) + str(dow) + ".sp3"
        outpath_opera = os.path.join(outpath, filename)

    elif outname == "auto_new_cnv":
        log.error("not implemented yet !!!!!")
        raise Exception

    f = open(outpath_opera, "w+")
    f.write(final_str)


def write_clk(
    df_clk_in, outpath, outname=None, prefix="orb", header="", output_std_values=False
):
    """
    Write a SP3 Clock file from an Clock DataFrame

    Parameters
    ----------
    df_clk_in : DataFrame
        Input Clock DataFrame.
    outpath : str
        The output path of the file (see also outname).
    outname : None or str, optional
        None = outpath is the full path (directory + filename) of the output.
        A string = a manual name for the file.
        'auto_old_cnv' = automatically generate the filename (old convention)
        'auto_new_cnv' = automatically generate the filename (new convention)
        The default is None.
    prefix : str, optional
        the output 3-char. name of the AC. The default is 'orb'.
    header : str, optional
        A string describing the clk file header. The default is "".
    output_std_values : bool, optional
        Add observation sigmas as the last column. The default is False.

    Returns
    -------
    The string containing the formatted clock data.
    """

    HEAD = header
    row_str_stk = []

    if output_std_values:
        row_str_proto = "{:2} {:4} {:4d} {:02d} {:02d} {:02d} {:02d} {:9.6f} {:2d}   {:19.12e} {:19.12e}"
    else:
        row_str_proto = (
            "{:2} {:4} {:4d} {:02d} {:02d} {:02d} {:02d} {:9.6f} {:2d}   {:19.12e}"
        )

    for irow, row in df_clk_in.iterrows():

        if output_std_values:
            one_or_two = 2
            row_str = row_str_proto.format(
                row["type"],
                row["name"],
                int(row["year"]),
                int(row["month"]),
                int(row["day"]),
                int(row["hour"]),
                int(row["minute"]),
                int(row["second"]),
                one_or_two,
                row["bias"],
                row["sigma"],
            )
        else:
            one_or_two = 1
            row_str = row_str_proto.format(
                row["type"],
                row["name"],
                int(row["year"]),
                int(row["month"]),
                int(row["day"]),
                int(row["hour"]),
                int(row["minute"]),
                int(row["second"]),
                one_or_two,
                row["bias"],
            )
        row_str_stk.append(row_str)

    ## Add EOF
    row_str_stk.append("EOF")

    corpse = "\n".join(row_str_stk)

    out = (
        HEAD
        + "                                                            END OF HEADER\n"
        + corpse
    )

    ### Manage the file path
    prefix_opera = prefix

    if not outname:
        outpath_opera = outpath
    elif outname == "auto_old_cnv":
        start_dt = dt.datetime(
            int(df_clk_in.iloc[0]["year"]),
            int(df_clk_in.iloc[0]["month"]),
            int(df_clk_in.iloc[0]["day"]),
        )
        week, dow = conv.dt2gpstime(start_dt)
        filename = prefix_opera + str(week) + str(dow) + ".clk"
        outpath_opera = os.path.join(outpath, filename)

    elif outname == "auto_new_cnv":
        log.error("not implemented yet !!!!!")
        raise Exception

    else:
        outpath_opera = os.path.join(outpath, outname)

    out = out

    with open(outpath_opera, "w+") as Fout:
        Fout.write(out)
        Fout.close()

    return out


def ine_block_mono(sat, dt_in, extra_intrvl_strt=0.1, extra_intrvl_end=0.4, step=300):
    """
    Write an EPOS INE block
    """

    fields = [
        "orb____1",
        "orb____2",
        "orb____3",
        "orb____4",
        "orb____5",
        "orb____6",
        "orb___db",
        "orb_s2db",
        "orb_c2db",
        "orb_s4db",
        "orb_c4db",
        "orb___yb",
        "orb___xb",
        "orb_sixb",
        "orb_coxb",
        "orb___cr",
    ]

    mjd = np.floor(conv.dt2mjd(dt_in))
    mjd_strt = mjd - extra_intrvl_strt
    mjd_end = mjd + extra_intrvl_end + 1

    lines = []

    l1 = " sat_nr  : " + sat + "\n"
    l2 = " stepsize: {:3}  {:6.2f}\n".format(sat, step)

    lines.append(l1)
    lines.append(l2)

    for field in fields:
        line = " {:}: {:3}  0.000000000000000E+00 {:11.5f} {:11.5f}\n".format(
            field, sat, mjd_strt, mjd_end
        )
        lines.append(line)

    lines.append(" end_sat\n")

    str_out = "".join(lines)

    return str_out


def write_ine_dummy_file(
    Sat_list,
    dt_in,
    extra_intrvl_strt=0.1,
    extra_intrvl_end=0.4,
    step=300,
    out_file_path=None,
):
    """
    Write an EPOS INE dummy (empty values) file
    """

    Lines = []

    mjd = np.floor(conv.dt2mjd(dt_in))
    mjd_strt = mjd - extra_intrvl_strt
    mjd_end = mjd + extra_intrvl_end + 1

    datestr = conv.dt2str(dt.datetime.now(), str_format="%Y/%m/%d %H:%M:%S")

    mjd_strt_deci = mjd_strt - np.floor(mjd_strt)

    head_proto = """%=INE 1.00 {:} NEWSE=INE+ORBCOR                                                                                 
+global
 day_info: 
 epoch   :                            {:5}  {:16.14f}
 interval:                            {:11.5f} {:11.5f}
 stepsize:      {:6.2f}
-global
+initial_orbit
"""
    head = head_proto.format(datestr, int(mjd), 0, mjd_strt, mjd_end, step)

    Lines.append(head)

    for sat in Sat_list:
        Lines.append(
            "******************************************************************\n"
        )
        sat_str = ine_block_mono(sat, dt_in, extra_intrvl_strt, extra_intrvl_end, step)
        Lines.append(sat_str)
        Lines.append(
            "******************************************************************\n"
        )

    str_end = """-initial_orbit
%ENDINE
"""

    Lines.append(str_end)

    str_out = "".join(Lines)

    if out_file_path:
        with open(out_file_path, "w") as f:
            f.write(str_out)
            f.close()

    return str_out


def sp3_overlap_creator(
    ac_list,
    dir_in,
    dir_out,
    suffix_out_input=None,
    overlap_size=7200,
    force=False,
    manage_missing_sats="exclude_missing_epoch",
    eliminate_null_sat=True,
    severe=False,
    separated_systems_export=False,
    first_date=None,
    end_date=None,
    # new_naming = False,
    exclude_bad_epoch=True,
    sys=None,
    new_name=False,
):
    """
    Generate an SP3 Orbit file with overlap based on the SP3s of the
    days before and after

    Parameters
    ----------
    ac_list : list
        3-character codes of the ACs.
    dir_in : str
        where the input sp3 are.
    dir_out : str
         where the output sp3 will be outputed.
    suffix_out_input : str, optional
        last char of the 3-char. code. if None, then it is the same as input.
    overlap_size : int, optional
        Overlapsize. The default is 7200.
    force : True, optional
        force overwrite. The default is False.
    manage_missing_sats : str, optional
        'exclude_missing_day' : generate a file with only the common sat
        between the 3 days. Thus, exclude the missing sats for a complete day\n
        'exclude_missing_epoch' : generate a file with only sat with full epochs\n
        'extrapolate' : extrapolate the missing sats based on the first/last epoch\n
        The default is 'exclude_missing_epoch'.
    eliminate_null_sat : bool, optional
        eliminate null sat. The default is True.
    severe : bool, optional
        raise an exception if problem. The default is False.
    separated_systems_export : bool, optional
        export different sp3 for different system. The default is False.
    first_date : datetime, optional
        exclude SP3 before this epoch
    end_date :datetime, optional
        exclude SP3 after this epoch
    exclude_bad_epoch : bool, optional
        remove bad epoch (usually filled with 99999999.9999999 or 0.00000000)
    sys : string, optional
        if just to keep one system (e.g.: G)

    Returns
    -------
    None.

    Note
    ----
    start/end date are not implemented
    the force option skips existing files
    not implemented for new names
    """

    Dict_Lfiles_ac = dict()

    for ac in ac_list:
        Dict_Lfiles_ac[ac] = []
        lfile = Dict_Lfiles_ac[ac]

        if new_name:
            lfile = operational.find_igs_products_files(
                dir_in,
                ["sp3"],
                ac_list,
                first_date,
                date_end=end_date,
                severe=False,
                recursive_search=True,
                regex_old_naming=True,
                regex_new_naming=True,
                regex_igs_tfcc_naming=False,
                compressed="incl",
            )

        else:
            # extlist = ["sp3","SP3","sp3.gz","SP3.gz"]
            extlist = ["sp3", "sp3.gz", "eph", "SP3", "SP3.gz"]
            for ext in extlist:
                lfile = lfile + utils.find_recursive(dir_in, "*" + ac + "*" + ext)

        log.info("Nb of SP3 found for %s %s", ac, len(lfile))

        if not suffix_out_input:
            suffix_out = ac
        else:
            suffix_out = (
                ac[:2] + suffix_out_input
            )  ## GM  2021-09-06 this is not good for the new namning convention

        d = []
        wwwwd = []

        new_name_list = []

        regex_suffix = "[0-9]{4}_[0-9]{2}[A-Z]_[0-9]{2}[A-Z]_ORB.SP3"

        for sp3 in lfile:
            # wwwwd_str = os.path.basename(sp3)[3:8]
            # d.append(conv.gpstime2dt(int(wwwwd_str[:4]),int(wwwwd_str[4:])))
            extension = sp3[-3:]
            dat = conv.sp3name2dt(sp3)
            d.append(dat)

            if re.search(regex_suffix, sp3):
                new_name_list.append(True)
            else:
                new_name_list.append(False)

        #        ## just a check for the files list. May be excluded in the future.
        #        for item in lfile:
        #            if 'COD0MGXFIN_20200050000_01D_05M_ORB.SP3' in item:
        #                print(lfile.index(item))

        for dat, newnamebool in zip(
            d[1:-1], new_name_list[1:-1]
        ):  ####if selection manuel, zip > 2lists !!!
            #        for dat in d[875:876]: ####if selection manuel, zip > 2lists !!!

            dummy = 1
            try:
                log.info("*********** %s %s", ac, dat)

                if first_date:
                    if dat < first_date:
                        log.info("SKIP date", dat)
                        continue
                if end_date:
                    if dat > end_date:
                        log.info("SKIP date", dat)
                        continue

                wwwwd_str = conv.dt_2_sp3_datestr(dat).zfill(5)

                dat_bef = dat - dt.timedelta(days=1)
                dat_aft = dat + dt.timedelta(days=1)

                wwwwd_str_bef = utils.join_improved(
                    "", *conv.dt2gpstime(dat_bef)
                ).zfill(5)
                wwwwd_str_aft = utils.join_improved(
                    "", *conv.dt2gpstime(dat_aft)
                ).zfill(5)

                ###### check if exists
                dir_out_wk = os.path.join(dir_out, "wk" + str(wwwwd_str)[:4])
                utils.create_dir(dir_out_wk)
                fil_out = dir_out_wk + "/" + suffix_out + wwwwd_str + ".sp3"

                if not force and os.path.isfile(fil_out):
                    log.info("0)) " + fil_out + " exists, skipping...")
                    continue

                ### *************** STEP 1 ***************
                log.info("1)) Search for the days before/after")
                log.info("1)) %s %s", dat_bef, dat_aft)

                ## GM 2021-09-06 in case of the new naming conventation this is step needs to be done:
                ## not the best way to handle this, since it is hardcoded and the step needs to be given in a "if"
                ## maybe try to find a better way with regex!
                # if new_naming:
                #     if ac in ['WUM','GRG','SHA']:
                #         step_str = '15'
                #     else:
                #         step_str = '05'

                #     year,day = conv.dt_to_doy(conv.gpstime2dt(int(wwwwd_str[:4]),int(wwwwd_str[-1])))
                #     p1    = utils.find_regex_in_list(str(year)+str(day).zfill(3)  + "0000_01D_"+step_str+"M_ORB.SP3",lfile,True)

                #     year_bef,day_bef = conv.dt_to_doy(conv.gpstime2dt(int(wwwwd_str_bef[:4]),int(wwwwd_str_bef[-1])))
                #     p_bef = utils.find_regex_in_list(str(year_bef)+str(day_bef).zfill(3)  + "0000_01D_"+step_str+"M_ORB.SP3",lfile,True)

                #     year_aft,day_aft = conv.dt_to_doy(conv.gpstime2dt(int(wwwwd_str_aft[:4]),int(wwwwd_str_aft[-1])))
                #     p_aft = utils.find_regex_in_list(str(year_aft)+str(day_aft).zfill(3)  + "0000_01D_"+step_str+"M_ORB.SP3",lfile,True)

                # if re.search(regex_suffix,)
                if newnamebool:
                    day, year = conv.dt2doy_year(dat)
                    regex_prefix = str(year) + str(day).zfill(3)
                    p1 = utils.find_regex_in_list(
                        regex_prefix + regex_suffix, lfile, True
                    )

                    day_bef, year_bef = conv.dt2doy_year(dat_bef)
                    regex_prefix_bef = str(year_bef) + str(day_bef).zfill(3)
                    p_bef = utils.find_regex_in_list(
                        regex_prefix_bef + regex_suffix, lfile, True
                    )

                    day_aft, year_aft = conv.dt2doy_year(dat_aft)
                    regex_prefix_aft = str(year_aft) + str(day_aft).zfill(3)
                    p_aft = utils.find_regex_in_list(
                        regex_prefix_aft + regex_suffix, lfile, True
                    )

                else:
                    if extension == "eph":
                        p1 = utils.find_regex_in_list(wwwwd_str + ".eph", lfile, True)
                        p_bef = utils.find_regex_in_list(
                            wwwwd_str_bef + ".eph", lfile, True
                        )
                        p_aft = utils.find_regex_in_list(
                            wwwwd_str_aft + ".eph", lfile, True
                        )
                    else:
                        p1 = utils.find_regex_in_list(wwwwd_str + ".sp3", lfile, True)
                        p_bef = utils.find_regex_in_list(
                            wwwwd_str_bef + ".sp3", lfile, True
                        )
                        p_aft = utils.find_regex_in_list(
                            wwwwd_str_aft + ".sp3", lfile, True
                        )

                log.info("1)) Files found for the days before/after")
                log.info("0b) %s", p_bef)
                log.info("01) %s", p1)
                log.info("0a) %s", p_aft)

                if not p1 or not p_bef or not p_aft:
                    log.error("with day %s", dat)
                    continue

                sp3 = files_rw.read_sp3(p1)
                sp3_bef = files_rw.read_sp3(p_bef)
                sp3_aft = files_rw.read_sp3(p_aft)

                ### Filtering to keep p only
                sp3 = sp3[sp3.type == "p"]
                sp3_bef = sp3_bef[sp3_bef.type == "p"]
                sp3_aft = sp3_aft[sp3_aft.type == "p"]

                if sys:
                    sp3 = sp3[sp3.sys == sys]
                    sp3_bef = sp3_bef[sp3_bef.sys == sys]
                    sp3_aft = sp3_aft[sp3_aft.sys == sys]

                sp3_bef = sp3_bef[sp3_bef["epoch"] < sp3["epoch"].min()]
                sp3_aft = sp3_aft[sp3_aft["epoch"] > sp3["epoch"].max()]

                sp3concat = pd.concat((sp3_bef, sp3, sp3_aft))

                dat_filter_bef = dat - dt.timedelta(seconds=overlap_size)
                dat_filter_aft = (
                    dat + dt.timedelta(seconds=overlap_size) + dt.timedelta(days=1)
                )

                ### *************** STEP 2 ***************
                log.info("2)) dates of the overlap period before/after")
                log.info("2)) %s %s", dat_filter_bef, dat_filter_aft)

                ### *************** STEP 3 ***************
                log.info("3)) Dates of: SP3 concatenated, before, current, after")
                log.info(
                    "3)) %s %s", sp3concat["epoch"].min(), sp3concat["epoch"].max()
                )
                log.info("3b) %s %s", sp3_bef["epoch"].min(), sp3_bef["epoch"].max())
                log.info("31) %s %s", sp3["epoch"].min(), sp3["epoch"].max())
                log.info("3a) %s %s", sp3_aft["epoch"].min(), sp3_aft["epoch"].max())

                sp3concat = sp3concat[
                    (sp3concat["epoch"] >= dat_filter_bef)
                    & (sp3concat["epoch"] <= dat_filter_aft)
                ]

                if exclude_bad_epoch:
                    good_epochs_bool_999 = sp3concat[["x", "y", "z"]] < 999999.0
                    good_epochs_bool_000 = np.logical_not(
                        np.isclose(sp3concat[["x", "y", "z"]], 0.0)
                    )

                    good_epochs_bool = np.logical_and(
                        good_epochs_bool_999, good_epochs_bool_000
                    )
                    good_epochs_bool = np.all(good_epochs_bool, axis=1)

                    sp3concat = sp3concat[good_epochs_bool]

                ########## HERE WE MANAGE THE MISSING SATS
                if manage_missing_sats == "exclude_missing_day":
                    log.info("4))", "remove missing sats -- day")
                    common_sats = (
                        set(sp3_bef["prn"])
                        .intersection(set(sp3["prn"]))
                        .intersection(set(sp3_aft["prn"]))
                    )
                    sp3concat = sp3concat[sp3concat["prn"].isin(common_sats)]

                elif manage_missing_sats == "exclude_missing_epoch":
                    log.info("4))", "remove missing sats -- epoch")
                    nepoc = len(sp3concat["epoch"].unique())
                    sp3concat_satgrp = sp3concat.groupby("prn")

                    all_sats = sp3concat["prn"].unique()
                    good_sats = sp3concat_satgrp.count() == nepoc

                    ###### good_sats = good_sats.reset_index()["sat"]
                    ## we get the good sats based one column containing a boolean
                    ## because of the test just before (abitrarily epoch column)
                    ## and after get the corresponding good sats names
                    good_sats = good_sats[good_sats["epoch"]].reset_index()["prn"]

                    bad_sats = list(set(all_sats) - set(good_sats))
                    log.info("excluded bad sats: %s", bad_sats)

                    sp3concat = sp3concat[sp3concat["prn"].isin(good_sats)]

                elif manage_missing_sats == "extrapolate":
                    log.info("4)) extrapolate missing sats ")
                    for iovl, SP3_ovl in enumerate((sp3_bef, sp3_aft)):
                        if iovl == 0:
                            backward = True
                            forward = False
                            backfor = "backward"
                        elif iovl == 1:
                            backward = False
                            forward = True
                            backfor = "forward"

                        sats = set(sp3["prn"])
                        sats_ovl = set(SP3_ovl["prn"])

                        sats_miss = sats.difference(sats_ovl)
                        if not sats_miss:
                            continue
                        log.info(
                            "4a) extrapolate missing sats %s %s", backfor, sats_miss
                        )

                        sp3extrapo_in = sp3concat[sp3concat["prn"].isin(sats_miss)]

                        # step = utils.most_common(sp3concat["epoch"].diff().dropna())
                        # step = step.astype('timedelta64[s]').astype(np.int32)
                        step = 900
                        # print(step)

                        # print("sp3extrapo_in",sp3extrapo_in)

                        sp3extrapo = reffram.extrapolate_sp3_data_frame(
                            sp3extrapo_in,
                            step=step,
                            n_step=int(overlap_size / step),
                            backward=backward,
                            forward=forward,
                            until_backward=dat_filter_bef,
                            until_forward=dat_filter_aft,
                            return_all=False,
                        )

                        sp3concat = pd.concat((sp3concat, sp3extrapo))
                        log.info(sp3extrapo)

                else:
                    log.error("check manage_missing_sats value")
                    raise Exception

                if eliminate_null_sat:
                    good_sats = []
                    for sat in sp3concat["prn"].unique():
                        xyzvals = sp3concat[sp3concat["prn"] == sat][
                            ["x", "y", "z"]
                        ].sum(axis=1)

                        v = np.sum(np.isclose(xyzvals, 0)) / len(xyzvals)

                        if v < 0.50:
                            good_sats.append(sat)
                        else:
                            log.info("6) eliminate because null position %s", sat)

                    sp3concat = sp3concat[sp3concat["prn"].isin(good_sats)]

                ### *************** STEP 7 ***************
                log.info("7)) Start/End Epoch of the concatenated file ")
                log.info(
                    "7)) %s %s", sp3concat["epoch"].min(), sp3concat["epoch"].max()
                )

                #### All systems
                log.info("8)) outputed file")
                log.info(fil_out)
                write_sp3(sp3concat, fil_out)

                #### system separated
                if False:
                    for sys in sp3concat["sys"].unique():
                        try:
                            sp3concat_sys = sp3concat[sp3concat["sys"] == sys]
                            fil_out_sys = (
                                dir_out_wk
                                + "/"
                                + suffix_out[:2]
                                + sys.lower()
                                + wwwwd_str.zfill(5)
                                + ".sp3"
                            )
                            log.info("9)) outputed file")
                            log.info(fil_out_sys)
                            write_sp3(sp3concat_sys, fil_out_sys)
                        except:
                            continue

            except KeyboardInterrupt:
                raise KeyboardInterrupt

            except Exception as e:
                if severe:
                    log.error(e)
                    raise e
                else:
                    log.warning("Error %s but no severe mode, continue...", e)
                    continue

    """
    sort_wrt="site" or "site_num"
    
    soln_in_df
    use soln AND pt information in the input DataFrame
    """


def write_epos_sta_coords(
    df_inp,
    file_out,
    sort_wrt="site",
    no_time_limit_for_first_period=True,
    no_time_limit_for_last_period=True,
    soln_in_df=True,
    trf_name="xTRFnn",
):
    """
    Write an EPOS coordinate file

    Parameters
    ----------
    df_inp : DataFrame
        Input Orbit DataFrame.
    file_out : str
        The output path of the file.
    sort_wrt : bool, optional
        Sort the values with respect to a DF column.
        The default is "site".
    no_time_limit_for_first_period : bool, optional
        No time limit for the first period.
        The default is True.
    no_time_limit_for_last_period : bool, optional
        No time limit for the last period.
        The default is True.
    soln_in_df : bool, optional
        Soln in DF.
        The default is True.

    Returns
    -------
    None.

    """

    df_work = df_inp.sort_values([sort_wrt, "mjd_start"])

    stat_lines_blk_stk = []

    generic_header = """+info
 FLATTENING                  298.2550
 MAJOR_AXIS              6378140.0000
 REFERENCE_FRAME                IGS14
 NUMBER_OF_STATIONS             {:5d}
 REF_MJD                        {:5d}
-info
"""

    generic_header = generic_header.format(
        len(df_work["site_num"].unique()), int(utils.most_common(df_work["MJD_ref"]))
    )

    stat_lines_blk_stk.append(generic_header)

    stat_lines_blk_stk.append("+station_coordinates")

    for site in df_work[sort_wrt].unique():

        stat_lines_blk_stk.append(
            "*------------------------- ---- ----- -beg- -end- -**- ------------------------------------------------\n*"
        )

        df_site_block = df_work[df_work[sort_wrt] == site]

        df_site_block.reset_index(inplace=True)

        for i_l, (_, l) in enumerate(df_site_block.iterrows()):

            if soln_in_df:
                iope = int(l["soln"])
                pt = l["pt"]
            else:
                iope = i_l + 1
                pt = "A"

            if no_time_limit_for_first_period and i_l == 0:
                mjd_start = 0
            else:
                mjd_start = l["mjd_start"]

            if no_time_limit_for_last_period and (i_l + 1) == len(df_site_block):
                mjd_end = 0
            else:
                mjd_end = l["mjd_end"]

            line_site_fmt = " SITE            m {:4d}  {:1d} {:} {:5d} {:5d} {:5d} {:}   {:}  {:1d}      LOG_CAR       LOG_CAR"
            line_site_fmt = " SITE            m {:4d}  {:1d} {:} {:5d} {:5d} {:5d} {:}   {:}  {:1d}{:>13}       {:<13}"
            line_valu_fmt = " POS_VEL:XYZ     m {:4d}  {:1d} {:+15.4f} {:+15.4f} {:+15.4f}      {:+6.4f} {:+6.4f} {:+6.4f}"
            line_sigm_fmt = " SIG_PV_XYZ      m {:4d}  {:1d} {:+15.4f} {:+15.4f} {:+15.4f}      {:+6.4f} {:+6.4f} {:+6.4f}"

            line_site = line_site_fmt.format(
                int(l["site_num"]),
                int(iope),
                l["tecto_plate"].upper(),
                int(l["MJD_ref"]),
                int(mjd_start),
                int(mjd_end),
                l["site"],
                pt,
                int(iope),
                trf_name,
                trf_name,
            )

            line_valu = line_valu_fmt.format(
                int(l["site_num"]),
                int(iope),
                l["x"],
                l["y"],
                l["z"],
                l["Vx"],
                l["Vy"],
                l["Vz"],
            )

            line_sigm = line_sigm_fmt.format(
                int(l["site_num"]),
                int(iope),
                l["sx"],
                l["sy"],
                l["sz"],
                l["sVx"],
                l["sVy"],
                l["sVz"],
            )

            stat_lines_blk_stk.append(line_site)
            stat_lines_blk_stk.append(line_valu)
            stat_lines_blk_stk.append(line_sigm)
            stat_lines_blk_stk.append("*")

    stat_lines_blk_stk.append("-station_coordinates")

    final_str = "\n".join(stat_lines_blk_stk)

    with open(file_out, "w+") as f:
        f.write(final_str)

    return final_str


# def write_sndy_light_dat(ts_in,outdir,outprefix):
#     """ Not properly implemented """
#     fil = open(os.path.join(outdir,outprefix),'w+')
#     if isinstance(ts_in,TimeSeriePoint):
#         if ts_in.initype() == 'FLH':
#             for pt in ts_in.pts:
#                 lin = ' '.join([str(e) for e in [pt.F , pt.L , pt.H , pt.T , pt.sF , pt.sL , pt.sH ]])
#                 fil.write(lin + '\n')
#     elif isinstance(ts_in,TimeSerieObs):
#         if ts_in.typeobs == 'RPY':
#             for att in ts_in.obs:
#                 lin = ' '.join([str(e) for e in [att.R , att.P , att.Y , att.T , att.Q.w , att.Q.x , att.Q.y , att.Q.z ]])
#                 fil.write(lin + '\n')
#     fil.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.operational contains functions to manipulate
RINEX files (version 2, quite obsolate).

it can be imported directly with:
from geodezyx import operational

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
import glob

#### Import the logger
import logging
import os
import re
import shutil
import string
import subprocess

import dateutil
import hatanaka
import numpy as np
import pandas as pd

#### geodeZYX modules
from geodezyx import conv
from geodezyx import utils

log = logging.getLogger("geodezyx")


##########  END IMPORT  ##########


def rinex_table_from_list(
    rnxs_inp,
    site9_col=False,
    round_date=False,
    path_col=True,
    size_col=False,
    sort=True,
):
    """
    From a simple RINEX list, summarize the data in an ad-hoc DataFrame.

    Parameters
    ----------
    rnxs_inp : iterable or str
        If iterable, a list of RINEX files.
        If str, path of an input RINEX list as text file.
    site9_col : bool, optional
        If True, include a 'site9' column in the DataFrame. Defaults to False.
    round_date : bool, optional
        If True, round the dates to the nearest day. Defaults to False.
    path_col : bool, optional
        If True, include the 'path' column in the DataFrame. Defaults to True.
    size_col : bool, optional
        If True, include a 'size' column in the DataFrame.
        Corresponds to the size of the file in bytes.
        It can slow down the process if the list is long.
        Defaults to False.
    sort : bool, optional
        If True, sort the DataFrame by 'site4' or 'site9' (if site9_col is True) and 'date'.
         Defaults to True.

    Returns
    -------
    df : pandas.DataFrame
        A DataFrame with the RINEX info in it.
    """
    if utils.is_iterable(rnxs_inp):
        df = pd.DataFrame(rnxs_inp, columns=["path"])
    else:
        df = pd.read_csv(rnxs_inp, names=["path"])

    df["name"] = df["path"].apply(os.path.basename)
    df["site4"] = df["name"].str[:4].str.lower()

    if site9_col:
        df["site9"] = df["name"].str[:9].str.lower()
        # Set generic 9-char name if 4 char name file
        bool_new_name = df["name"].str.match(conv.rinex_regex_long_name())
        bool_old_name = np.logical_not(bool_new_name)
        site9_generic = df["site4"].str.upper() + "00XXX"
        df.loc[bool_old_name, "site9"] = site9_generic[bool_old_name]

    df["date"] = df["name"].apply(conv.rinexname2dt)
    if round_date:
        df["date"] = conv.round_dt(df["date"].values, "1D", True, "floor")
    doy_year = df["date"].apply(conv.dt2doy_year)
    df["doy"] = doy_year.apply(lambda x: x[0])
    df["year"] = doy_year.apply(lambda x: x[1])

    if size_col:
        size_path = lambda f: os.path.getsize(f) if os.path.isfile(f) else np.nan
        df["size"] = df["path"].apply(size_path)

    cols = df.columns.tolist()
    cols = cols[1:] + [cols[0]]
    df = df[cols]

    if not path_col:
        df.drop("path", axis=1, inplace=True)

    pd.options.display.max_info_columns = 150

    if sort:
        site_col = "site9" if site9_col else "site4"
        df.sort_values(by=[site_col, "date"], inplace=True)
        df.reset_index(drop=True, inplace=True)

    return df


def rinex_lists_diff(rnx_lis1, rnx_lis2, out_dir=None, out_name=None, site9_col=False):
    """
    Compare two lists of RINEX files and identify differences.

    Parameters
    ----------
    rnx_lis1 : list
        First list of RINEX file paths.
    rnx_lis2 : list
        Second list of RINEX file paths.
    out_dir : str, optional
        Directory to save the output files. If None, no files are saved.
    out_name : str, optional
        Base name for the output files. Defaults to 'rnx_diff' if not provided.
    site9_col : bool, optional
        If True, use 'site9' as the column name for site identification. Defaults to False.

    Returns
    -------
    diff1m2 : pandas.DataFrame
        DataFrame containing entries in rnx_lis1 but not in rnx_lis2.
    diff2m1 : pandas.DataFrame
        DataFrame containing entries in rnx_lis2 but not in rnx_lis1.
    intrsec : pandas.DataFrame
        DataFrame containing entries common to both rnx_lis1 and rnx_lis2.
    symdiff : pandas.DataFrame
        DataFrame containing entries that are in either rnx_lis1 or rnx_lis2 but not in both.
    """

    site_col = "site9" if site9_col else "site4"

    # Create DataFrames from the RINEX lists
    d1 = rinex_table_from_list(
        rnx_lis1, path_col=True, round_date=False, site9_col=site9_col
    )
    d2 = rinex_table_from_list(
        rnx_lis2, path_col=True, round_date=False, site9_col=site9_col
    )

    # Set the index to the site and date columns
    d1 = d1.set_index([site_col, "date"])
    d2 = d2.set_index([site_col, "date"])

    # Find the differences in the indices
    idx_diff1m2 = d1.index.difference(d2.index)
    idx_diff2m1 = d2.index.difference(d1.index)
    idx_intrsec = d1.index.intersection(d2.index)
    idx_symdiff = d1.index.symmetric_difference(d2.index)

    # Locate the differences in the DataFrames
    diff1m2 = d1.loc[idx_diff1m2]
    diff2m1 = d2.loc[idx_diff2m1]
    intrsec = d1.loc[idx_intrsec]
    d12 = pd.concat([d1, d2], axis=0)
    symdiff = d12.loc[idx_symdiff]

    diff1m2.sort_index(inplace=True)
    diff2m1.sort_index(inplace=True)
    intrsec.sort_index(inplace=True)
    symdiff.sort_index(inplace=True)

    res_dic = dict()
    res_dic["diff. 1-2"] = diff1m2
    res_dic["diff. 2-1"] = diff2m1
    res_dic["intersection"] = intrsec
    res_dic["sym. diff."] = intrsec

    # If an output directory is specified, save the differences to files
    if out_dir:
        if not out_name:
            out_name = ""

        # Save the paths of the differences to text files
        # Save the full DataFrames of the differences to CSV files
        for k, v in res_dic.items():
            if len(v) > 0:
                out_suffix = k.replace(".", "").replace("-", "_").replace(" ", "_")
                out_path = os.path.join(out_dir, out_name + "_" + out_suffix)
                log.info("Saving rinex %s to %s", k, out_path + ".txt/.csv")
                v["path"].to_csv(out_path + ".txt", index=False, header=False)
                v.to_csv(out_path + ".csv", index=False, header=True)

    return diff1m2, diff2m1, intrsec, symdiff


def rinex_sats_checker(p_rnx):
    """
    Check the consistency of a RINEX's sat list for each epoch

    Designed for checking the consistency of IPGP-OVPF corrupted RINEXs
    RINEX-2 only

    Parameters
    ----------
    p_rnx : str
        input RINEX path.

    Returns
    -------
    bad RINEX path or None.

    """

    import hatanaka

    ### This bloc must  be replaced by open_readlines_smart
    if p_rnx.endswith(".Z") or p_rnx.endswith(".gz"):
        O = hatanaka.decompress(p_rnx)
        O = str(O)
        L = O.split("\\n")
    else:
        F = open(p_rnx, "r+")
        L = F.readlines()

    iline_bloc = 0
    nlines_bloc = -1

    blocs_stack = []

    ###### This loop read all lines
    for l in L:
        re_epoch = "^ {1,2}([0-9]{1,2} * ){5}"
        # re_sat="[A-Z][0-9][0-9]"
        bool_epoch = re.search(re_epoch, l)
        # bool_sat=re.search(re_sat,l)

        ### we found an epoch line
        if bool_epoch:
            in_epoch = True
            nsat = int(l[30:32])
            iline_bloc = 0
            nlines_bloc = int(np.ceil(nsat / 12))
            line_bloc = []
            lineconcat = ""
            date = read_rnx_epoch_line(l, rnx2=True)

        ### we read the sat lines based on the number of sat
        if iline_bloc <= nlines_bloc:
            line_bloc.append(l[32:].strip())
            lineconcat = lineconcat + l[32:].strip()
            iline_bloc += 1

        ### we stack everything when the sat block is over
        if iline_bloc == nlines_bloc:
            in_epoch = False
            bloc_tuple = (date, nsat, lineconcat)
            blocs_stack.append(bloc_tuple)

    #### we do a DF with all the found line
    df = pd.DataFrame(blocs_stack)

    df[3] = None

    for irow, row in df.iterrows():
        nsat = str(row[1])
        satstr = row[2]
        re_sat = "([A-Z][0-9][0-9]){" + nsat + "}"
        bool_sat = re.search(re_sat, satstr)

        #### the the right number of sat is found True = Good
        if bool_sat:
            df.iloc[irow, 3] = True
        else:
            df.iloc[irow, 3] = False

    if df[3].sum() != len(df):
        return rinex_sats_checker
    else:
        return None


def read_rnx_epoch_line(line, rnx2=True):
    """
    read the epoch line of a RINEX (2 or 3)
    and extract it as a datetime

    Parameters
    ----------
    line : string
        input line, assuming it's a RINEX epoch line.
    rnx2 : TYPE, optional
        True if RINEX2, False if RINEX3/4.
        The default is True.

    Returns
    -------
    epochdt : datetime
        The epoch of the line.
    """

    if rnx2:
        strtidx = 0
    else:
        strtidx = 1
    fraw = line[strtidx:].split()  # fraw est la pour gerer les fracion de sec ...
    fraw = fraw[0:6]
    fraw = [float(e) for e in fraw]

    f = [int(e) for e in fraw]
    msec = fraw[5] - np.floor(fraw[5])
    msec = np.round(msec, 4)
    msec = int(msec * 10**6)
    f.append(msec)  # ajour des fractions de sec

    if rnx2:
        if f[0] < 50:
            f[0] = f[0] + 2000
        else:
            f[0] = f[0] + 1900

    if f[5] == 60:  # cas particulier rencontré dans des rinex avec T = 60sec
        rinex_60sec = True
        f[5] = 59
    else:
        rinex_60sec = False

    epochdt = dt.datetime(*f)

    if rinex_60sec:
        epochdt = epochdt + dt.timedelta(seconds=1)
    return epochdt


#  _____  _____ _   _ ________   __   _____       _ _ _
# |  __ \|_   _| \ | |  ____\ \ / /  / ____|     | (_) |
# | |__) | | | |  \| | |__   \ v /  | (___  _ __ | |_| |_ ___ _ __
# |  _  /  | | | . ` |  __|   > <    \___ \| '_ \| | | __/ _ \ '__|
# | | \ \ _| |_| |\  | |____ / . \   ____) | |_) | | | ||  __/ |
# |_|  \_\_____|_| \_|______/_/ \_\ |_____/| .__/|_|_|\__\___|_|
#                                          | |
#                                          |_|


def check_if_compressed_rinex(rinex_path):
    boolout = bool(re.search(r".*((d|o)\.(Z)|(gz))$", rinex_path))
    return boolout


def crz2rnx(rinex_path, outdir="", force=True, path_of_crz2rnx="CRZ2RNX", verbose=True):
    """assuming that CRZ2RNX is in the system PATH per default"""

    if not os.path.isfile(rinex_path):
        raise Exception(rinex_path + " dont exists !")

    if force:
        forcearg = "-f"
    else:
        forcearg = ""

    curdir = os.getcwd()
    command = path_of_crz2rnx + " -c " + forcearg + " " + rinex_path
    if outdir == "":
        outdir = os.path.dirname(rinex_path)

    out_rinex_name_splited = os.path.basename(rinex_path).split(".")

    if conv.rinex_regex_search_tester(rinex_path, short_name=False, long_name=True):
        out_rinex_name_splited = os.path.basename(rinex_path).split(".")
        out_rinex_name = out_rinex_name_splited[0]
        out_rinex_name = out_rinex_name + ".rnx"
        out_rinex_path = os.path.join(outdir, out_rinex_name)
    else:
        out_rinex_name_splited = os.path.basename(rinex_path).split(".")
        out_rinex_name = ".".join(out_rinex_name_splited[:-1])
        out_rinex_name = out_rinex_name[:-1] + "o"
        out_rinex_path = os.path.join(outdir, out_rinex_name)

    if os.path.isfile(out_rinex_path) and not force:
        log.warning(out_rinex_path + "already exists, skiping ...")
    else:
        os.chdir(outdir)
        # stream = os.popen(command)

        proc = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE, executable="/bin/bash"
        )
        status = proc.wait()

        if status != 0 or not os.path.isfile(out_rinex_path):
            log.error(out_rinex_path + " not created!")
        elif verbose:
            log.info(out_rinex_path + " created")
        else:
            pass
        #    print stream.read()
        os.chdir(curdir)
    return out_rinex_path


def rnx2crz(rinex_path, outdir="", force=True, path_of_rnx2crz="RNX2CRZ"):
    """assuming that RNX2CRZ is in the system PATH per default"""

    if not os.path.isfile(rinex_path):
        raise Exception(rinex_path + " dont exists !")

    if force:
        forcearg = "-f"
    else:
        forcearg = ""

    curdir = os.curdir
    command = path_of_rnx2crz + " " + forcearg + " " + rinex_path
    if outdir == "":
        outdir = os.path.dirname(rinex_path)

    out_rinex_name_o = os.path.basename(rinex_path)
    out_rinex_name_d_z = out_rinex_name_o[:-1] + "d.Z"
    out_rinex_path = os.path.join(outdir, out_rinex_name_d_z)

    if os.path.isfile(out_rinex_path) and not force:
        log.info(out_rinex_path + "already exists, skiping ...")
    else:
        os.chdir(outdir)
        stream = os.popen(command)
        log.info(command + " output in " + outdir)
        #    print stream.read()
    os.chdir(curdir)
    return out_rinex_path


def rinex_read_epoch(
    input_rinex_path_or_string,
    interval_out=False,
    add_tzinfo=False,
    out_array=True,
    out_index=False,
):
    """
    Read the epochs contained in a RINEX File. Can handle RINEX 2 and 3

    Parameters
    ----------
    input_rinex_path_or_string : see below
        input RINEX.
        can be the path of a RINEX file as string or as Path object,
        or directly the RINEX content as a string, bytes, StringIO object or a
        list of lines

    interval_out : bool, optional
        output also the intervals. The default is False.

    add_tzinfo : bool, optional
        add timezone information in the datetime's Epoches.
        The default is False.

    out_array : bool, optional
        output results as array.
        The default is True.

    out_index : bool, optional
        output also the index of the epoch line.
        The default is False.

    Returns
    -------
    array or list
        the epochs in the RINEX.
    """

    ##161019 : dirty copier coller de rinex start end
    epochs_list = []
    rinex_60sec = False

    try:
        input_rinex_path_or_string = hatanaka.decompress(input_rinex_path_or_string)
    except:
        pass

    rnx_lines = utils.open_readlines_smart(input_rinex_path_or_string)

    index_list = []
    for iline, line in enumerate(rnx_lines):
        epoch_rnx2 = re.search("^ {1,2}([0-9]{1,2} * ){5}", line)
        epoch_rnx3 = re.search("^>", line)

        if epoch_rnx2:
            try:
                epochdt = read_rnx_epoch_line(line, rnx2=True)
                epochs_list.append(epochdt)
                index_list.append(iline)
            except:
                log.error(line)

        elif epoch_rnx3:
            try:
                epochdt = read_rnx_epoch_line(line, rnx2=False)
                epochs_list.append(epochdt)
                index_list.append(iline)
            except:
                log.error(line)

    if add_tzinfo:
        epochs_list = [e.replace(tzinfo=dateutil.tz.tzutc()) for e in epochs_list]

    if out_index:
        output = [epochs_list, index_list]
    else:
        output = epochs_list

    if out_array:
        output = np.array(output).T

    return output


def same_day_rinex_check(rinex1, rinex2):
    if os.path.basename(rinex1)[4:7] == os.path.basename(rinex2)[4:7]:
        return True
    else:
        return False


def rinex_start_end(
    input_rinex_path,
    interval_out=False,
    add_tzinfo=False,
    verbose=True,
    safety_mode=True,
):
    """
    Return the first and the last epoch of a RINEX file
    (based on the actual content of the file, not the header)

    Can handle RINEX 2 and 3

    Parameters
    ----------
    input_rinex_path : TYPE
        path of the rinex file.
        can be the path of a RINEX or directly
        the RINEX content as a string

    interval_out : bool, optional
        output also the intervals. The default is False.

    add_tzinfo : bool, optional
        add timezone information in the datetime's Epoches.
        The default is False.

    verbose : bool, optional
        verbose mode. The default is True.

    safety_mode : TYPE, optional
        if the epoch reading fails (e.g. in case of a compressed RINEX)
        activate a reading of the header and the file name as backup.
        The default is True.

    Returns
    -------
    first_epoch , last_epoch , [interval]
        First, las epoches and interval if asked.

    """

    # une liste d'epochs en début et fin de fichier
    # => en trouver le min et le max
    # NB : FAIRE UN FONCTION READ EPOCH A L'OCCAZ
    # NBsuite : c'est fait au 161018 mais par contre c'est un dirty copier coller

    epochs_list = []
    head = utils.head(input_rinex_path, 1500)
    epochs_list_head = rinex_read_epoch(
        head, interval_out=interval_out, add_tzinfo=add_tzinfo, out_array=False
    )

    tail = utils.tail(input_rinex_path, 1500)
    epochs_list_tail = rinex_read_epoch(
        tail, interval_out=interval_out, add_tzinfo=add_tzinfo, out_array=False
    )

    epochs_list = epochs_list_head + epochs_list_tail
    
    if len(epochs_list) > 0:
        first_epoch = np.min(epochs_list)
        last_epoch = np.max(epochs_list)
    else:
        ## here is the save mode where everything can goes wrong ;)
        first_epoch = conv.rinexname2dt(input_rinex_path)
        if first_epoch:
            last_epoch = first_epoch + dt.timedelta(seconds=85399)
        else:
            ### here the return is None, your file is corrupted
            last_epoch = None
            return first_epoch, last_epoch

        ##### this legacy block if for short name only
        # alphabet = list(string.ascii_lowercase)
        # if os.path.basename(input_rinex_path)[7] in alphabet:
        #    last_epoch = first_epoch + dt.timedelta(hours=1)
        # else:
        #    last_epoch = first_epoch + dt.timedelta(hours=24,seconds=-1)

    if add_tzinfo:
        first_epoch = first_epoch.replace(tzinfo=dateutil.tz.tzutc())
        last_epoch = last_epoch.replace(tzinfo=dateutil.tz.tzutc())

    if verbose:
        log.debug("first & last epochs: %s %s", first_epoch, last_epoch)

    if not interval_out:
        return first_epoch, last_epoch
    else:
        interv_lis = np.diff(epochs_list)
        interv_lis = [e.seconds + e.microseconds * 10**-6 for e in interv_lis]
        interval = utils.most_common(interv_lis)
        log.debug("interval, last epoch: %s %s", interval, last_epoch)

        # return interv_lis , epochs_list
        return first_epoch, last_epoch, interval


def rinex_session_id(first_epoch, last_epoch, full_mode=False):
    """
    full_mode:
        gives the letter of the starting session  & the length in hour
        of the session
    """

    alphabet = list(string.ascii_lowercase)
    rinex_length = last_epoch - first_epoch
    if rinex_length.seconds > 86000:
        rnx_interval_ext = "0"
    else:
        if not full_mode:
            rnx_interval_ext = alphabet[first_epoch.hour]
        else:
            rnx_interval_ext = alphabet[first_epoch.hour] + str(
                int(rinex_length.seconds / 3600.0)
            )

    return rnx_interval_ext


# def rinex_spliter(input_rinex_path,output_directory,stat_out_name='',
#                  interval_size=24,compress=False,shift = 0):
#    """
#    if shift != 0:
#    the start/end of a session is shifted of shift minutes
#    """
#
#    if stat_out_name == '':
#        stat_out_name = os.path.basename(input_rinex_path)[0:4]
#
#    # check if the RINEX is compressed ...
#    bool_comp_rnx = check_if_compressed_rinex(input_rinex_path)
#    # ... if not crz2rnx !
#    if bool_comp_rnx:
#        input_rinex_path = crz2rnx(input_rinex_path)
#
#    inp_rinex_obj=open(input_rinex_path,'r+')
#    out_dir = output_directory # nom redondant mais j'ai la flemme d'aller corriger le nom de la variable
#
#    if not os.path.exists(out_dir):
#        os.makedirs(out_dir)
#    os.chdir(out_dir)
#
#    first_epoch , last_epoch = rinex_start_end(input_rinex_path)
#
#    # In this function, Date are truncated epochs
#    # (only the day if interval == 24 , + the hour else)
#    first_date = datetime.datetime(first_epoch.year,first_epoch.month,
#                                   first_epoch.day,first_epoch.hour)
#    last_date = datetime.datetime(last_epoch.year,last_epoch.month,
#                                  last_epoch.day,last_epoch.hour)+datetime.timedelta(hours=1)
#
#    print "first & last dates (truncated) : " , first_date , last_date
#
#    time_interval = genefun.get_interval(first_date,last_date,
#                                         datetime.timedelta(hours=interval_size))
#
#    alphabet = list(string.ascii_lowercase)
#
#    rinex_out_name_lis = []
#
#    for i,curr_date in enumerate(time_interval):
#        if interval_size == 24:
#            rnx_interval_ext = '0.'
#        else:
#            rnx_interval_ext = alphabet[curr_date.hour] + '.'
#
#        p = subprocess.Popen('',executable='/bin/bash', stdin=subprocess.PIPE , stdout=subprocess.PIPE , stderr=subprocess.PIPE)
#        command = 'teqc ' + '-O.mo ' + stat_out_name.upper() +' -st ' +  curr_date.strftime('%y%m%d%H%M%S') + ' +dh ' + str(interval_size) + ' ' + input_rinex_path
#        print command
#        rinex_out_name = stat_out_name + curr_date.strftime('%j') + rnx_interval_ext + curr_date.strftime('%y') + 'o'
#        err_log_name = curr_date.strftime('%j') + rnx_interval_ext + "err.log"
#        print rinex_out_name
#        stdout,stderr = p.communicate( command )
#        std_file = open(rinex_out_name, "w")
#        std_file.write(stdout)
#        std_file.close()
#        if stderr != '':
#            print err_log_name + " is not empty, must be checked !"
#            print stderr
#            err_file = open(err_log_name, "w")
#            err_file.write(stderr)
#            err_file.close()
#
#        rnx_splt_path = os.path.join(out_dir,rinex_out_name)
#        if compress:
#            rinex_out_final = rnx2crz(rnx_splt_path)
#            os.remove(rnx_splt_path)
#        else:
#            rinex_out_final = rnx_splt_path
#
#        rinex_out_name_lis.append(rinex_out_final)
#
#    return rinex_out_name_lis


def rinex_spliter(
    input_rinex_path,
    output_directory,
    stat_out_name="",
    interval_size=24,
    compress=False,
    shift=0,
    inclusive=False,
    teqc_cmd="teqc",
):
    """
    if shift != 0:
    the start/end of a session is shifted of shift minutes

    inclusive:
    delta of exaclty interval_size => add the 1st epoch of the next sess
    not inclusive:
    delta of interval_size - 1s
    """
    """
    Split RINEX file using teqc

    Parameters
    ----------
    input_rinex_path : str
        Path of the input RINEX

    output_directory : str
        Description param2
        
    stat_out_name : str
        Station name for the output RINEX
        
    interval_size : int
        Size of the splitted interval
    
    compress : bool
        compress the outputed RINEX    

    interval_size : int
        Size of the splitted interval 
        
        
    FINIR CE HEADER !!!!
                
    Returns
    -------
    out1 : float or int or str or dict or n-tuple or bool or list or numpy.array
        Description out1
    
    out2 : float or int or str or dict or n-tuple or bool or list or numpy.array
        Description out2
        
    Note
    ----
    Misc. Notes

    Source
    ------
    www.forum-source.com
    
    Examples
    --------
    >> answer
    42    
    """

    if stat_out_name == "":
        stat_out_name = os.path.basename(input_rinex_path)[0:4]

    # check if the RINEX is compressed ...
    bool_comp_rnx = check_if_compressed_rinex(input_rinex_path)

    # ... if not crz2rnx !
    if bool_comp_rnx:
        input_rinex_path = crz2rnx(input_rinex_path)

    inp_rinex_obj = open(input_rinex_path, "r+")
    out_dir = output_directory  # nom redondant mais j'ai la flemme d'aller corriger le nom de la variable

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    os.chdir(out_dir)

    first_epoch, last_epoch = rinex_start_end(input_rinex_path)

    # In this function, Date are truncated epochs
    # (only the day if interval == 24 , + the hour else)
    first_date = dt.datetime(
        first_epoch.year, first_epoch.month, first_epoch.day, first_epoch.hour
    )
    last_date = dt.datetime(
        last_epoch.year, last_epoch.month, last_epoch.day, last_epoch.hour
    ) + dt.timedelta(hours=1)

    log.info("first & last dates (truncated): %s %s", first_date, last_date)

    time_interval = utils.get_interval(
        first_date, last_date, dt.timedelta(hours=interval_size)
    )

    alphabet = list(string.ascii_lowercase)

    rinex_out_name_lis = []

    for i, curr_date in enumerate(time_interval):

        if shift != 0:
            log.warning("shifted mode on, be careful !!!")

        if bool(shift) and i != 0:
            curr_date = curr_date + dt.timedelta(minutes=shift)
            interval_size_ope = interval_size
        elif bool(shift) and i == 0:
            interval_size_ope = interval_size + float(shift) / 60.0
        else:
            interval_size_ope = interval_size

        if not inclusive:
            interval_size_ope = interval_size_ope - 1 / 3600.0

        if not bool(shift):
            if interval_size == 24:
                rnx_interval_ext = "0."
            else:
                rnx_interval_ext = alphabet[curr_date.hour] + "."
        else:
            if interval_size == 24:
                rnx_interval_ext = "0."
            else:
                rnx_interval_ext = alphabet[curr_date.hour] + "."

        p = subprocess.Popen(
            "",
            executable="/bin/bash",
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        # - 1/3600. # one_sec2h
        command = (
            teqc_cmd
            + " -O.mo "
            + stat_out_name.upper()
            + " -st "
            + curr_date.strftime("%y%m%d%H%M%S")
            + " +dh "
            + str(interval_size_ope)
            + " "
            + input_rinex_path
        )
        log.info(command)
        rinex_out_name = (
            stat_out_name
            + curr_date.strftime("%j")
            + rnx_interval_ext
            + curr_date.strftime("%y")
            + "o"
        )
        err_log_name = curr_date.strftime("%j") + rnx_interval_ext + "err.log"
        log.info(rinex_out_name)
        stdout, stderr = p.communicate(command.encode())
        std_file = open(rinex_out_name, "w")
        std_file.write(stdout.decode("utf-8"))
        std_file.close()
        if stderr:
            log.warning(err_log_name + " is not empty, must be checked !")
            log.info(stderr.decode("utf-8"))
            err_file = open(err_log_name, "w")
            err_file.write(stderr.decode("utf-8"))
            err_file.close()

        rnx_splt_path = os.path.join(out_dir, rinex_out_name)
        if compress:
            rinex_out_final = rnx2crz(rnx_splt_path)
            os.remove(rnx_splt_path)
        else:
            rinex_out_final = rnx_splt_path

        rinex_out_name_lis.append(rinex_out_final)

    return rinex_out_name_lis


def rinex_spliter_gfzrnx(
    input_rinex_path,
    output_directory,
    stat_out_name="",
    interval_size=86400,
    shift=0,
    inclusive=False,
    gfzrnx_cmd="GFZRNX",
    output_name="::RX3::",
    custom_cmds="",
):
    interval_size_hour = interval_size / 3600.0

    if stat_out_name == "":
        stat_out_name = os.path.basename(input_rinex_path)[0:4]

    # check if the RINEX is compressed ...
    bool_comp_rnx = check_if_compressed_rinex(input_rinex_path)

    # ... if not crz2rnx !
    if bool_comp_rnx:
        input_rinex_path = crz2rnx(input_rinex_path)

    inp_rinex_obj = open(input_rinex_path, "r+")
    out_dir = output_directory  # nom redondant mais j'ai la flemme d'aller corriger le nom de la variable

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    os.chdir(out_dir)

    first_epoch, last_epoch = rinex_start_end(input_rinex_path)

    # In this function, Date are truncated epochs
    # (only the day if interval == 24 , + the hour else)
    first_date = dt.datetime(
        first_epoch.year, first_epoch.month, first_epoch.day, first_epoch.hour
    )
    last_date = dt.datetime(
        last_epoch.year, last_epoch.month, last_epoch.day, last_epoch.hour
    ) + dt.timedelta(hours=1)

    log.info("first & last dates (truncated): %s %s", first_date, last_date)

    time_interval = utils.get_interval(
        first_date, last_date, dt.timedelta(hours=interval_size_hour)
    )

    alphabet = list(string.ascii_lowercase)

    rinex_out_name_lis = []

    for i, curr_date in enumerate(time_interval):

        if shift != 0:
            log.warning("shifted mode on, be careful !!!")

        if bool(shift) and i != 0:
            curr_date = curr_date + dt.timedelta(seconds=shift)
            interval_size_ope = interval_size
        # elif bool(shift) and i == 0 :
        #    interval_size_ope = interval_size + float(shift) / 60.
        else:
            interval_size_ope = interval_size

        if not inclusive:
            interval_size_ope = interval_size_ope - 1.0

        interval_size_ope = int(np.floor(interval_size_ope))

        if not bool(shift):
            if interval_size == 24:
                rnx_interval_ext = "0."
            else:
                rnx_interval_ext = alphabet[curr_date.hour] + "."
        else:
            if interval_size == 24:
                rnx_interval_ext = "0."
            else:
                rnx_interval_ext = alphabet[curr_date.hour] + "."

        p = subprocess.Popen(
            "",
            executable="/bin/bash",
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # - 1/3600. # one_sec2h
        command = (
            gfzrnx_cmd
            + " -finp "
            + input_rinex_path
            + " -fout "
            + os.path.join(output_directory, output_name)
            + " -site "
            + stat_out_name.upper()
            + " -epo_beg "
            + curr_date.strftime("%Y%m%d_%H%M%S")
            + " --duration "
            + str(interval_size_ope)
            + " "
            + custom_cmds
        )
        log.info(command)

        # rinex_out_name = stat_out_name + curr_date.strftime('%j') + rnx_interval_ext + curr_date.strftime('%y') + 'o'
        std_log_name = curr_date.strftime("%j") + rnx_interval_ext + "err.log"
        err_log_name = curr_date.strftime("%j") + rnx_interval_ext + "err.log"

        stdout, stderr = p.communicate(command.encode())

        std_file = open(std_log_name, "w")
        std_file.write(stdout.decode("utf-8"))
        std_file.close()

        if stderr:
            log.info(err_log_name + " output:")
            log.info(stderr.decode("utf-8"))
            err_file = open(err_log_name, "w")
            err_file.write(stderr.decode("utf-8"))
            err_file.close()

        ### get the latest files
        list_of_files = glob.glob(
            output_directory + "/*"
        )  # * means all if need specific format then *.csv
        latest_file = sorted(list_of_files, key=os.path.getctime)[
            -3:
        ]  ### -2 and -1 are the logs

        latest_file = [e for e in latest_file if ".rnx" in e]

        if len(latest_file) == 0:
            latest_file = ""
            rnx_splt_path = ""
        else:
            latest_file = latest_file[0]
            rnx_splt_path = os.path.join(out_dir, latest_file)

        rinex_out_name_lis.append(rnx_splt_path)


def teqc_qc(rinex_path, quick_mode=False, optional_args=""):
    """
    quick mode : reduced qc and no summary file written
    """
    if quick_mode:
        qcmode = "+qcq"
    else:
        qcmode = "+qc"

    # check if the RINEX is compressed ...
    bool_comp_rnx = check_if_compressed_rinex(rinex_path)

    # ... if not crz2rnx !
    if bool_comp_rnx:
        rinex_path_work = crz2rnx(rinex_path)
    else:
        rinex_path_work = rinex_path

    kommand = " ".join(("teqc", qcmode, optional_args, rinex_path))

    log.info(kommand)

    proc = subprocess.Popen(
        kommand, shell=True, stdout=subprocess.PIPE, executable="/bin/bash"
    )
    status = proc.wait()

    if bool_comp_rnx:
        os.remove(rinex_path_work)

    if status:
        log.error("crash for " + rinex_path + ", code " + str(proc.poll()))
        return None

    output = proc.stdout.read()

    return output.decode("ASCII")


def rinex_renamer(input_rinex_path, output_directory, stat_out_name="", remove=False):
    if stat_out_name == "":
        stat_out_name = os.path.basename(input_rinex_path)[0:4]

    stat_out_name = stat_out_name.lower()

    inp_rinex_obj = open(input_rinex_path, "r+")
    out_dir = output_directory  # nom redondant mais j'ai la flemme d'aller corriger le nom de la variable

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    os.chdir(out_dir)

    first_epoch, last_epoch = rinex_start_end(input_rinex_path)
    rnx_interval_ext = rinex_session_id(first_epoch, last_epoch) + "."
    rinex_out_name = (
        stat_out_name
        + first_epoch.strftime("%j")
        + rnx_interval_ext
        + first_epoch.strftime("%y")
        + "o"
    )

    log.info(rinex_out_name)

    output_rinex_path = os.path.join(out_dir, rinex_out_name)

    if input_rinex_path != output_rinex_path:
        log.info("copy of " + input_rinex_path + " to " + output_rinex_path)
        shutil.copy(input_rinex_path, output_rinex_path)

        if remove and os.isfile(output_rinex_path):
            log.info("removing ", input_rinex_path)
            os.remove(input_rinex_path)
    else:
        log.info(input_rinex_path)
        log.info("and " + output_rinex_path + " are the same file")
        log.info("nothing's done ...")

    return output_rinex_path


# def rinex_renamer_gfz_odc_2_igs(rinex_in,output_dir=None):



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 19/04/2025 17:32:02

@author: psakic
"""

import os
import shutil
import subprocess
import time
import yaml
import glob

import geodezyx.operational.gins_runner.gins_common as gynscmn




#   ____  _     _   _                                    __     _
#  / __ \| |   | | | |                                  / _|   | |
# | |  | | | __| | | |     ___  __ _  __ _  ___ _   _  | |_ ___| |_ ___
# | |  | | |/ _` | | |    / _ \/ _` |/ _` |/ __| | | | |  _/ __| __/ __|
# | |__| | | (_| | | |___|  __/ (_| | (_| | (__| |_| | | || (__| |_\__ \
#  \____/|_|\__,_| |______\___|\__, |\__,_|\___|\__, | |_| \___|\__|___/
#                               __/ |            __/ |
#                              |___/            |___/


############### OLD FCTS
def write_oclo_file(station_file, oceanload_out_file, fes_yyyy=2004):
    temp_cmd_fil = os.path.join(os.path.dirname(oceanload_out_file), "oclo.cmd.tmp")
    temp_cmd_filobj = open(temp_cmd_fil, "w")

    temp_cmd_filobj.write(station_file + "\n")
    temp_cmd_filobj.write(oceanload_out_file + "\n")
    if fes_yyyy == 2012:
        exe_loadoce_cmd = "exe_loadoce_fes2012"
    elif fes_yyyy == 2004:
        exe_loadoce_cmd = "exe_loadoce"
    else:
        exe_loadoce_cmd = "exe_loadoce"

    p = subprocess.Popen(
        exe_loadoce_cmd + " < " + temp_cmd_fil,
        shell=True,
        executable="/bin/bash",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    #    os.remove(temp_cmd_fil)
    temp_cmd_filobj.close()

    if True:
        return oceanload_out_file
    else:
        return ""



def dirs_copy_generik_2_working(
    director_name_prefix, director_generik_path, temp_data_folder=None
):

    if not temp_data_folder:
        temp_data_folder = gynscmn.get_temp_data_gins_path()

    shutil.copy(director_generik_path, temp_data_folder)
    director_generik_path_tmp = os.path.join(
        temp_data_folder, os.path.basename(director_generik_path)
    )
    director_generik_path_out = os.path.join(
        temp_data_folder, director_name_prefix + ".yml"
    )
    os.rename(director_generik_path_tmp, director_generik_path_out)

    return director_generik_path_out


def get_director_list(wildcard_dir):
    """with a wildcard (e.g. 'GWADA_MK2*') and return a list of corresponding
    directors found in gin/data/directeur folder"""
    gins_path = gynscmn.get_gin_path()
    di_run_lis = [
        os.path.basename(e)
        for e in glob.glob(os.path.join(gins_path, "data", "directeur", wildcard_dir))
    ]
    return di_run_lis


def sort_by_stations(archive_path, wildcard, i):
    """i is the indice of the first character of the station name in
    eg : i = 18 for filename KARIB_MK3_FLH_v2__bara_22282_2011_003

    archive path is ABSOLUTE b.c. it can be outside of the gins folder"""

    path = os.path.join(archive_path, wildcard)
    filis = glob.glob(path)

    firstfil = os.path.basename(filis[0])
    print("sure the stat name is good ? ")
    print("e.g.", firstfil[i : i + 4], "for", firstfil)
    print("(you have 5sec to abort if not)")
    time.sleep(5)

    for f in filis:
        stat = os.path.basename(f)[i : i + 4]
        stat_up = stat.upper()
        archiv = os.path.join(archive_path, stat_up)
        if not os.path.exists(archiv):
            os.makedirs(archiv)
        shutil.move(f, archiv)

    return None


def merge_yaml(yaml1, yaml2, yaml_out=None):

    dic1 = yaml.load(open(yaml1))
    dic2 = yaml.load(open(yaml2))

    def merge(dic1, dic2):
        if isinstance(dic1, dict) and isinstance(dic2, dict):
            for k, v in dic2.items():
                if k not in dic1:
                    dic1[k] = v
                else:
                    dic1[k] = merge(dic1[k], v)
        return dic1

    dic3 = merge(dic1, dic2)

    if yaml_out:
        with open(yaml_out, "w+") as outfile:
            outfile.write(yaml.dump(dic3, default_flow_style=False))

    return dic3


def check_gins_exe(streamin, director_name):
    """
    Check the execution status of a GINS director.

    Parameters
    ----------
    streamin : file-like object
        The input stream to read the execution output from.
    director_name : str
        The name of the director being checked.

    Returns
    -------
    bool
        True if the execution was successful, False otherwise.
    """
    if "Exécution terminée du fichier" in streamin.read():
        print("INFO : happy end for " + director_name + " :)")
        return True
    else:
        print("WARN : bad end for " + director_name + " :(")
        return False

def run_dirs_kwwrap(kwarg):
    run_directors(**kwarg)
    return None


def run_dirs_multi(
    dir_paths_inp,
    nprocs=4,
    opts_gins_pc="",
    opts_gins_90="",
    version="OPERA",
    cmd_mode="exe_gins_dir",
    force=False,
):
    kwargs_lis = []
    for dirr in list(sorted(dir_paths_inp)):
        kwargs = dict()
        kwargs["dir_paths_inp"] = dirr
        kwargs["opts_gins_pc"] = opts_gins_pc
        kwargs["opts_gins_90"] = opts_gins_90
        kwargs["version"] = version
        kwargs["cmd_mode"] = cmd_mode
        kwargs["force"] = force
        kwargs_lis.append(kwargs)

    pool = mp.Pool(processes=nprocs)
    # res_raw = [pool.apply(run_dirs_kwwrap, args=(x,)) for x in kwargs_lis]
    res_raw = pool.map(run_dirs_kwwrap, kwargs_lis, chunksize=1)

    return None



def run_dirs_multislots_custom(
    director_lis,
    slots_lis=["", "U", "L", "R"],
    opts_gins_pc="",
    opts_gins_90="",
    version="OPERA",
    mode="ginspc",
):
    if not type(director_lis) is list:
        log.error("director_lis in input is not a list !!!")
        return None
    log.info("TOTAL NB OF DIRECTORS : %i", len(director_lis))

    if type(slots_lis) is int:
        slots_lis = [""] * slots_lis

    chunk = utils.chunkIt(director_lis, len(slots_lis))
    pool_size = len(slots_lis)
    pool = mp.Pool(processes=pool_size)
    ZIP = list(
        zip(
            chunk,
            [slot + opts_gins_pc for slot in slots_lis],
            [opts_gins_90] * len(slots_lis),
            [version] * len(slots_lis),
            [mode] * len(slots_lis),
        )
    )
    pool.map(run_director_list_wrap, ZIP)
    return None


def smart_directors_to_run(wildcard_dir="", full_path_out=True):
    """smart runner check directors who worked, and thus give a list of directors
    whitout those who worked

    listing and directeur folders are inspected automatically"""

    gins_path = gynscmn.get_gin_path(True)

    # list of corresponding directors
    dirpath = os.path.join(gins_path, "data", "directeur", wildcard_dir)
    dirlis = glob.glob(dirpath)
    dirlis = [os.path.basename(e) for e in dirlis]

    # list of corresponding already lauched listings
    lipath = os.path.join(gins_path, "batch", "listing", wildcard_dir)
    lilis = glob.glob(lipath)

    lilis = [os.path.basename(e) for e in lilis]
    # must treat new and old independently
    lilis_new = [e for e in lilis if ".yml" in e]
    lilis_old = [e for e in lilis if not ".yml" in e]
    # 1) treating OLD case
    # MAKING LISTS OF TUPLES (PREFIX , <gins or prepars>)
    li_tuple_lis_old = [(e.split(".")[0], e.split(".")[-1]) for e in lilis_old]
    li_finished_lis_old = [e[0] for e in li_tuple_lis_old if "gins" in e[1]]

    # 2) treating NEW case
    # MAKING LISTS OF TUPLES (PREFIX , <gins or prepars>)
    li_tuple_lis_new = [
        (e.split(".")[0] + "." + e.split(".")[1], e.split(".")[-1]) for e in lilis_new
    ]
    li_finished_lis_new = [e[0] for e in li_tuple_lis_new if "gins" in e[1]]

    li_finished_lis = li_finished_lis_old + li_finished_lis_new

    # director to run are those in the dir list minus those which worked
    di_run_lis = list(set(dirlis) - set(li_finished_lis))

    print("for wildcard :", wildcard_dir)
    print("input directors found in the data/dir. folder     :", len(dirlis))
    print("FINISHED listings found in the batch/list. folder :", len(li_finished_lis))
    print("directors who need to be launched                 :", len(di_run_lis))

    di_run_lis = sorted(di_run_lis)

    if full_path_out:
        dirpath = os.path.join(gynscmn.get_gin_path(True), "data", "directeur")
        di_run_lis = [os.path.join(dirpath, d) for d in di_run_lis]

    return di_run_lis


def smart_listing_archive(
    wildcard_dir,
    gins_main_archive,
    gins_anex_archive,
    prepars_archive,
    director_archive,
):
    """for each listing corresponding to the wildcard :
    if it's a prepars => go to the prepars_archive
    if it's a gins without duplicate  => go to the gins_main_archive
    if it's a gins with duplcates => one goes to gins_main_archive
                                     the others in gins_anex_archive"""

    listing_path = os.path.join(
        gynscmn.get_gin_path(), "gin", "batch", "listing", wildcard_dir
    )
    listing_lis = [e for e in glob.glob(listing_path)]

    prepars_lis = [e for e in listing_lis if "prepars" in e]
    gins_lis = [e for e in listing_lis if "gins" in e]

    dirpath = os.path.join(
        gynscmn.get_gin_path(True), "data", "directeur", wildcard_dir
    )
    dirlis = [e for e in glob.glob(dirpath)]

    # searching for doublons
    # 2 runs of the same day/stat => keeping the first one
    # Only new case (.yml) is treated ...

    gins_prefix_lis = [os.path.basename(e).split(".")[0] for e in gins_lis]
    #    gins_tuple_lis  = [(e.split('.')[0] , e.split('.')[2] ,e.split('.')[3]) for e in gins_lis]

    count_gins_prefix = collections.Counter(gins_prefix_lis).most_common()
    multi_gins_prefix = [elt for elt, count in count_gins_prefix if count > 1]
    simpl_gins_prefix = [elt for elt, count in count_gins_prefix if count == 1]

    simpl_gins_fullpath = []
    for s in simpl_gins_prefix:
        simpl_gins_fullpath = simpl_gins_fullpath + [e for e in gins_lis if s in e]

    multi_gins_fullpath_main = []
    multi_gins_fullpath_anex = []

    for m in multi_gins_prefix:
        multi_gins_fullpath_temp = [e for e in gins_lis if m in e]
        multi_gins_fullpath_main.append(multi_gins_fullpath_temp[0])
        multi_gins_fullpath_anex = (
            multi_gins_fullpath_anex + multi_gins_fullpath_temp[1:]
        )

    for directory in [
        gins_main_archive,
        gins_anex_archive,
        prepars_archive,
        director_archive,
    ]:
        if not os.path.exists(directory):
            os.makedirs(directory)

    # MOVING
    for p in prepars_lis:
        shutil.move(p, prepars_archive)
    for g in simpl_gins_fullpath:
        shutil.move(g, gins_main_archive)
    for g in multi_gins_fullpath_main:
        shutil.move(g, gins_main_archive)
    for g in multi_gins_fullpath_anex:
        shutil.move(g, gins_anex_archive)
    for d in dirlis:
        shutil.move(d, director_archive)

    return None


def export_results_gins_listing(
    listings_list_in, outpath, static_or_kinematic="kine", outprefix="", coordtype="FLH"
):
    from geodezyx import time_series
    if len(listings_list_in) == 0:
        print("ERR : export_results_gins_listing : listings list is empty ...")
        return None
    if static_or_kinematic == "stat":
        ts = files_rw.read_gins_multi_raw_listings(listings_list_in)
    elif static_or_kinematic == "kine":
        tslist = []
        for lising in listings_list_in:
            tslist.append(files_rw.read_gins(lising))
        ts = time_series.merge_ts(tslist)
    time_series.export_ts(ts, outpath, coordtype, outprefix)
    return outpath


def run_director_list_wrap(tupinp):
    run_directors(*tupinp)
    return None




#  ______                _   _                                                             _
# |  ____|              | | (_)                                                           | |
# | |__ _   _ _ __   ___| |_ _  ___  _ __     __ _ _ __ __ ___   _____ _   _  __ _ _ __ __| |
# |  __| | | | '_ \ / __| __| |/ _ \| '_ \   / _` | '__/ _` \ \ / / _ \ | | |/ _` | '__/ _` |
# | |  | |_| | | | | (__| |_| | (_) | | | | | (_| | | | (_| |\ v /  __/ |_| | (_| | | | (_| |
# |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|  \__, |_|  \__,_| \_/ \___|\__, |\__,_|_|  \__,_|
#                                             __/ |                     __/ |
#                                            |___/                     |___/


#    return dic1

# yaml1    = '/home/psakicki/GINS/gin/data/EXE_PPP/DIR_REF_PPP.yml'
# yaml2    = '/home/psakicki/GINS/gin/TP/TP_RELAX/DIR_MC0.yml'
# yaml_out = '/home/psakicki/GINS/gin/TP/TP_RELAX/DIR_MC0_perso2.yml'
#
# merge_yaml(yaml1,yaml2,yaml_out)


###### FUNCTIONS GRAVEYARD ########

# def get_rinex_list(
#     parent_folder,
#     specific_stats=[],
#     invert=False,
#     compressed=True,
#     start=dt.datetime(1980, 1, 1),
#     end=dt.datetime(2099, 1, 1),
# ):
#     """return :
#         all RINEXs found in a parent folder and his subfolders
#         (compressed or not)
#
#     parent_folder :
#         can be a the path of the partent folder (the rinex archive path)
#         but also a RINEX list already found
#         (modification of 170524 to gain speed)
#
#     specific_stats :
#         MUST BE a list ['STA1','STA2','STA3']
#         So, if only one elt in specific_stats, use a syntax as : ['STA1']
#
#     invert :
#         False = keeping the specific stats
#         True  = removing the specific stats
#         or for all stations, leave a empty
#         tuple in specific_stats
#
#     NB : the end date is included
#     """
#     if type(parent_folder) is str:
#         wholefilelist = []
#         for root, dirs, files in os.walk(parent_folder, topdown=False):
#             for name in files:
#                 wholefilelist.append(os.path.join(root, name))
#
#         wholefilelist = list(set(wholefilelist))
#
#         if compressed:
#             rnxregex = operational.rinex_regex()
#         else:
#             rnxregex = operational.rinex_regex(False)
#
#         rinexfilelist = [fil for fil in wholefilelist if re.search(rnxregex, fil)]
#     elif utils.is_iterable(parent_folder):
#         # is parent_folder is a list containing already found RINEXs
#         rinexfilelist = parent_folder
#
#     specific_stats = [e.lower() for e in specific_stats]
#
#     if specific_stats != []:
#         if not invert:
#             goodrnxfilelist = [
#                 fil
#                 for fil in rinexfilelist
#                 if os.path.basename(fil)[0:4] in specific_stats
#             ]
#         else:
#             goodrnxfilelist = [
#                 fil
#                 for fil in rinexfilelist
#                 if os.path.basename(fil)[0:4] not in specific_stats
#             ]
#     else:
#         goodrnxfilelist = rinexfilelist
#
#     goodrnxfilelist = sorted(goodrnxfilelist)
#
#     if goodrnxfilelist == []:
#         print(
#             "WARN : get_rinex_list : no RINEX found, check the path of the parent folder !"
#         )
#     else:
#         print("INFO : get_rinex_list :", len(goodrnxfilelist), "RINEXs found")
#
#     goodrnxfilelist_date = []
#     end = end + dt.timedelta(seconds=86399)
#     for rnx in goodrnxfilelist:
#         dtrnx = conv.rinexname2dt(os.path.basename(rnx))
#         if start <= dtrnx <= end:
#             goodrnxfilelist_date.append(rnx)
#     if len(goodrnxfilelist_date) != len(goodrnxfilelist):
#         dif = len(goodrnxfilelist) - len(goodrnxfilelist_date)
#         print(
#             "INFO : get_rinex_list :",
#             dif,
#             "RINEXs removed bc. not in the date interval",
#         )
#
#     if goodrnxfilelist_date == []:
#         print("WARN : get_rinex_list : all RINEXs removed bc. of date interval")
#
#     return goodrnxfilelist_date



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .geo_files_converter_lib import *
from .read_athmo import *
from .read_coords_misc import *
from .read_coords_time_series import *
from .read_eop import *
from .read_geo_files_misc import *
from .read_gnss_prods import *
from .read_gnss_qc import *
from .read_igs_combi import *
from .read_logsheets import *
from .read_rinex import *
from .read_rinex_nav import *
from .read_sinex import *
from .read_slr import *
from .write_geo_files import *
from .write_rinex import *
from .read_antex import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 27/08/2025 22:34:53

@author: psakic
"""

import datetime as dt
import os
from io import BytesIO
import numpy as np
import pandas as pd

def read_rinex_nav(fn, writeh5=None, version=2):
    """
    Based on Michael Hirsch readRinexNav function
    Manage RINEX3
    http://gage14.upc.es/gLAB/HTML/GPS_Navigation_Rinex_v2.11.html
    """

    with open(fn, "r") as f:
        # find end of header, which has non-constant length
        while True:

            line = f.readline()

            if "RINEX VERSION" in line:
                if float(line.split()[0]) < 3:
                    version = 2
                else:
                    version = 3

            if "END OF HEADER" in line:
                break

            if version == 2:
                startcol = 3  # column where numerical data starts
                nfloat = 19  # number of text elements per float data number
                nline = 7  # number of lines per record
            elif version == 3:
                startcol = 5  # column where numerical data starts
                nfloat = 19  # number of text elements per float data number
                nline = 7  # number of lines per record

        # handle frame by frame
        sv = []
        epoch = []
        raws = ""

        while True:
            headln = f.readline()
            if not headln:
                break
            # handle the header
            if version == 2:
                sv.append(headln[:2])
                year = int(headln[2:5])
                if 80 <= year <= 99:
                    year += 1900
                elif year < 80:  # good till year 2180
                    year += 2000

                epoch.append(
                    dt.datetime(
                        year=year,
                        month=int(headln[5:8]),
                        day=int(headln[8:11]),
                        hour=int(headln[11:14]),
                        minute=int(headln[14:17]),
                        second=int(headln[17:20]),
                        microsecond=int(headln[21]) * 100000,
                    )
                )

                raw = (
                    headln[22:].rstrip()
                    + "".join(
                        f.readline()[startcol:].rstrip() for _ in range(nline - 1)
                    )
                    + f.readline()[startcol:79].rstrip()
                )

            if version == 3:
                fields = headln[:23].split(" ")
                sv.append(fields[0])
                year = int(fields[1])

                epoch.append(
                    dt.datetime(
                        year=year,
                        month=int(fields[2]),
                        day=int(fields[3]),
                        hour=int(fields[4]),
                        minute=int(fields[5]),
                        second=int(fields[6]),
                    )
                )

                """
                now get the data.
                Use rstrip() to chomp newlines consistently on Windows and Python 2.7/3.4
                Specifically [:-1] doesn't work consistently as .rstrip() does here.
                """
                raw = (
                    headln[23:].rstrip()
                    + " "
                    + "".join(
                        f.readline()[startcol:].rstrip() + " " for _ in range(nline - 1)
                    )
                    + f.readline()[startcol:80].rstrip()
                )

            raws += raw + "\n"

    raws = raws.replace("D", "E")

    strio = BytesIO(raws.encode())
    darr = np.genfromtxt(strio, delimiter=nfloat)

    nav = pd.DataFrame(
        darr,
        epoch,
        [
            "SVclockBias",
            "SVclockDrift",
            "SVclockDriftRate",
            "IODE",
            "Crs",
            "DeltaN",
            "M0",
            "Cuc",
            "Eccentricity",
            "Cus",
            "sqrtA",
            "TimeEph",
            "Cic",
            "OMEGA",
            "CIS",
            "Io",
            "Crc",
            "omega",
            "OMEGA DOT",
            "IDOT",
            "CodesL2",
            "GPSWeek",
            "L2Pflag",
            "SVacc",
            "SVhealth",
            "TGD",
            "IODC",
            "TransTime",
            "FitIntvl",
            "spare1",
            "spare2",
        ],
    )

    if version == 3:
        const = [e[0] for e in sv]
        svni = [int(e[1:]) for e in sv]
        nav["sys"] = pd.Series(np.array(const), index=nav.index)
        nav["svni"] = pd.Series(np.array(svni), index=nav.index)
    elif version == 2:
        rinexnav_type = os.path.basename(fn)[-1]
        if rinexnav_type == "n":
            nav["sys"] = pd.Series(["G"] * len(nav.index), index=nav.index)
        else:
            nav["sys"] = pd.Series(
                [rinexnav_type.upper()] * len(nav.index), index=nav.index
            )
        nav["svni"] = pd.Series(np.array(sv), index=nav.index)

    if False:
        h5fn = fn.with_suffix(".h5")
        print(("saving NAV data to {}".format(h5fn)))
        nav.to_hdf(h5fn, key="NAV", mode="a", complevel=6, append=False)

    return nav



# https://chat.deepseek.com/a/chat/s/b979bb2d-15d3-41e7-880c-736e686e6161

import numpy as np
import pandas as pd
import xarray as xr
from datetime import datetime
import uuid

def create_insitu_tac_netcdf(output_path):
    """
    Create a prototype Copernicus Marine In Situ TAC NetCDF file.
    
    Parameters:
    - output_path: str, path where to save the NetCDF file
    """
    
    # Create sample data
    num_obs = 100
    time = pd.date_range(start="2023-01-01", periods=num_obs, freq="H")
    latitude = np.random.uniform(-90, 90, num_obs)
    longitude = np.random.uniform(-180, 180, num_obs)
    depth = np.random.uniform(0, 5000, num_obs)
    temperature = np.random.uniform(-2, 30, num_obs)
    salinity = np.random.uniform(30, 38, num_obs)
    
    # Create platform IDs and types
    platforms = [f"platform_{i:03d}" for i in range(num_obs)]
    platform_types = ["buoy"] * num_obs
    
    # Create QC flags (0=no QC, 1=good, 2=probably good, etc.)
    qc_flags = np.random.choice([0, 1, 2, 3, 4], size=num_obs)
    
    # Create xarray Dataset
    ds = xr.Dataset(
        {
            "LATITUDE": (["OBS"], latitude),
            "LONGITUDE": (["OBS"], longitude),
            "DEPH": (["OBS"], depth),
            "TEMP": (["OBS"], temperature),
            "PSAL": (["OBS"], salinity),
            "POSITION_QC": (["OBS"], qc_flags),
            "TEMP_QC": (["OBS"], qc_flags),
            "PSAL_QC": (["OBS"], qc_flags),
            "PLATFORM_NUMBER": (["OBS"], platforms),
            "PLATFORM_TYPE": (["OBS"], platform_types),
        },
        coords={
            "OBS": np.arange(num_obs),
            "TIME": ("OBS", time),
        }
    )
    
    # Add global attributes (following Copernicus conventions)
    ds.attrs = {
        "title": "Copernicus Marine In Situ TAC prototype",
        "institution": "Your Institution",
        "source": "In situ observations",
        "history": f"{datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')} - File created",
        "references": "http://marine.copernicus.eu",
        "Conventions": "CF-1.8, Copernicus-InSituTAC-2.0",
        "featureType": "trajectory",
        "id": str(uuid.uuid4()),
        "data_mode": "D",  # D for delayed mode, R for real-time
        "format_version": "2.0",
        "netcdf_version": "4",
        "naming_authority": "Copernicus",
        "cdm_data_type": "Trajectory",
        "comment": "Prototype file - not for operational use",
        "geospatial_lat_min": str(latitude.min()),
        "geospatial_lat_max": str(latitude.max()),
        "geospatial_lon_min": str(longitude.min()),
        "geospatial_lon_max": str(longitude.max()),
        "time_coverage_start": str(time[0]),
        "time_coverage_end": str(time[-1]),
        "date_created": datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
    }
    
    # Add variable attributes (modified to avoid calendar conflict)
    ds["TIME"].attrs = {
        "standard_name": "time",
        "long_name": "Time",
        "axis": "T",
    }
    
    ds["LATITUDE"].attrs = {
        "standard_name": "latitude",
        "long_name": "Latitude",
        "units": "degrees_north",
        "valid_min": "-90",
        "valid_max": "90",
        "axis": "Y",
    }
    
    ds["LONGITUDE"].attrs = {
        "standard_name": "longitude",
        "long_name": "Longitude",
        "units": "degrees_east",
        "valid_min": "-180",
        "valid_max": "180",
        "axis": "X",
    }
    
    ds["DEPH"].attrs = {
        "standard_name": "depth",
        "long_name": "Depth",
        "units": "meters",
        "positive": "down",
        "axis": "Z",
    }
    
    ds["TEMP"].attrs = {
        "standard_name": "sea_water_temperature",
        "long_name": "Sea temperature",
        "units": "degree_Celsius",
        "valid_min": "-5",
        "valid_max": "40",
    }
    
    ds["PSAL"].attrs = {
        "standard_name": "sea_water_salinity",
        "long_name": "Practical salinity",
        "units": "1",
        "valid_min": "0",
        "valid_max": "50",
    }
    
    # QC flags attributes
    for var in ["POSITION_QC", "TEMP_QC", "PSAL_QC"]:
        ds[var].attrs = {
            "long_name": f"Quality flag for {var.split('_')[0]}",
            "conventions": "Copernicus Marine In Situ reference table 2",
            "flag_values": "0,1,2,3,4,5,6,7,8,9",
            "flag_meanings": (
                "no_qc_performed good_data probably_good_data probably_bad_data bad_data "
                "changed_value value_below_detection value_in_excess interpolated_value missing_value"
            ),
        }
    
    ds["PLATFORM_NUMBER"].attrs = {
        "long_name": "Platform unique identifier",
        "conventions": "WMO float identifier : A9IIIII",
    }
    
    ds["PLATFORM_TYPE"].attrs = {
        "long_name": "Platform type",
        "conventions": "Copernicus Marine In Situ reference table 25",
    }
    
    # Save to NetCDF with proper encoding
    encoding = {
        "TIME": {"dtype": "float64", "_FillValue": None, "calendar": "gregorian"},
        "LATITUDE": {"dtype": "float32", "_FillValue": None},
        "LONGITUDE": {"dtype": "float32", "_FillValue": None},
        "DEPH": {"dtype": "float32", "_FillValue": None},
        "TEMP": {"dtype": "float32", "_FillValue": -9999.0},
        "PSAL": {"dtype": "float32", "_FillValue": -9999.0},
    }
    
    ds.to_netcdf(output_path, encoding=encoding, format="NETCDF4")
    print(f"File saved to {output_path}")

# Example usage
if __name__ == "__main__":
    create_insitu_tac_netcdf("copernicus_insitu_tac_prototype.nc")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb  9 07:43:47 2024

@author: snahmani
"""


# -*- coding: utf-8 -*-
"""
"""
import numpy as np
import pandas as pd

import geodezyx.gnss_edu as gnss_edu


def gpt3_5_fast_readGrid(filename ='gpt3_5.grd'):
# =============================================================================  
#     gpt3_5_fast_readGrid.py
#     This routine reads the grid 'gpt3_5.grd' and overgives the respective
#     parameters in a cell. It must be placed ahead of the for loop, which runs
#    through all observations in order to save a huge time amount. The related
#    function "gpt3_5_fast.py" must then replace the default
#   "gpt3_5.m" at the respective location in the text.
    
    
#    File created by Daniel Landskron, 2017-04-12
    
#    File Converted to python script by Zohreh Adavi zohreh.adavi@tuwien.ac.at, 2023-02-20
# =============================================================================  
    
# read gridfile

    
    file=open(filename)
    C = pd.read_csv(file ,skiprows=[i for i in range(0,1)],sep=r'\s+',skipfooter =0, skip_blank_lines=True,header = None, index_col= False )
    file.close()   


    
    p_grid = C.loc[:,np.arange(2,7)]
    
    T_grid = C.loc[:,np.arange(7,12)]
    
    Q_grid = C.loc[:,np.arange(12,17)] / 1000
    
    dT_grid= C.loc[:,np.arange(17,22)] / 1000
    
    u_grid = C.loc[:,22]
    
    Hs_grid = C.loc[:,23]
    
    ah_grid = C.loc[:,np.arange(24,29)] / 1000 
    
    aw_grid = C.loc[:,np.arange(29,34)] / 1000 
    
    la_grid = C.loc[:,np.arange(34,39)]
    
    Tm_grid = C.loc[:,np.arange(39,44)]
    
    Gn_h_grid = C.loc[:,np.arange(44,49)] / 100000
    
    Ge_h_grid = C.loc[:,np.arange(49,54)] / 100000
    
    Gn_w_grid = C.loc[:,np.arange(54,59)] / 100000 
    
    Ge_w_grid = C.loc[:,np.arange(59,64)] / 100000
    
    # combine all data to on cell grid
    gpt3_grid={}
    gpt3_grid[0]=p_grid
    gpt3_grid[1]=T_grid
    gpt3_grid[2]=Q_grid
    gpt3_grid[3]=dT_grid
    gpt3_grid[4]=u_grid
    gpt3_grid[5]=Hs_grid
    gpt3_grid[6]=ah_grid
    gpt3_grid[7]=aw_grid
    gpt3_grid[8]=la_grid
    gpt3_grid[9]=Tm_grid
    gpt3_grid[10]=Gn_h_grid
    gpt3_grid[11]=Ge_h_grid
    gpt3_grid[12]=Gn_w_grid
    gpt3_grid[13]=Ge_w_grid
            
    return gpt3_grid



    
def gpt3_5_fast(mjd = None,lat = None,lon = None,h_ell = None,it = None,grid = None): 
    
#    (c) Department of Geodesy and Geoinformation, Vienna University of Technology, 2017
    
# =============================================================================  
#     This subroutine determines pressure, temperature, temperature lapse rate,
#     mean temperature of the water vapor, water vapour pressure, hydrostatic
#     and wet mapping function coefficients ah and aw, water vapour decrease
#     factor, geoid undulation and empirical tropospheric gradients for
#     specific sites near the earth's surface. All output values are valid for
#     the specified ellipsoidal height h_ell.
#     GPT3_5 is based on a 5°x5° external grid file ('gpt3_5.grd') with mean
#     values as well as sine and cosine amplitudes for the annual and
#     semiannual variation of the coefficients.
#     This file goes together with 'gpt3_5_fast_readGrid.m', which reads the
#     grid and saves it in cell structures, what, especially for large numbers
#     of stations, significantly reduces runtime compared to gpt3_5.m.
#     gpt3_5_fast_readGrid' must be placed ahead of the for loop, in which
#     gpt3_5_fast is positioned.
    
    
#     Reference:
#     D. Landskron, J. Böhm (2018), VMF3/GPT3: Refined Discrete and Empirical Troposphere Mapping Functions,
#     J Geod (2018) 92: 349., doi: 10.1007/s00190-017-1066-2.
#     Download at: https://link.springer.com/content/pdf/10.1007#2Fs00190-017-1066-2.pdf
    
    
#     Input parameters:
#     -------------------    
#     mjd:   modified Julian date (scalar, only one epoch per call is possible)
#     lat:   ellipsoidal latitude in radians [-pi/2:+pi/2] (vector)
#     lon:   longitude in radians [-pi:pi] or [0:2pi] (vector)
#     h_ell: ellipsoidal height in m (vector)
#     it:    case 1: no time variation but static quantities
#     case 0: with time variation (annual and semiannual terms)
    
#     Output parameters:
#     -------------------
#     p:    pressure in hPa (vector)
#     T:    temperature in degrees Celsius (vector)
#     dT:   temperature lapse rate in degrees per km (vector)
#     Tm:   mean temperature weighted with the water vapor in degrees Kelvin (vector)
#     e:    water vapour pressure in hPa (vector)
#     ah:   hydrostatic mapping function coefficient (VMF3) (vector)
#     aw:   wet mapping function coefficient (VMF3) (vector)
#     la:   water vapour decrease factor (vector)
#     undu: geoid undulation in m (vector)
#     Gn_h: hydrostatic north gradient in m (vector)
#     Ge_h: hydrostatic east gradient in m (vector)
#     Gn_w: wet north gradient in m (vector)
#     Ge_w: wet east gradient in m (vector)
    
#     File created by Daniel Landskron, 2017-04-12
#     File Converted to python script by Zohreh Adavi zohreh.adavi@tuwien.ac.at, 2023-02-20
# =============================================================================  
    
# convert mjd to doy
    
    hour = int(np.floor((mjd - int(np.floor(mjd))) * 24))
    
    minu = int(np.floor((((mjd - int(np.floor(mjd))) * 24) - hour) * 60))
    
    sec = (((((mjd - int(np.floor(mjd))) * 24) - hour) * 60) - minu) * 60
    
    # change secs, min hour whose sec==60
    
    
    if sec==60:
        minu=minu+1
        sec=0
        
    if minu==60:
        hour=hour+1
        minu=0
    # calc jd (yet wrong for hour==24)
    jd = mjd + 2400000.5
    # if hr==24, correct jd and set hour==0
    if hour==24:
        jd=jd+1
        hour=0
    
    # integer julian date
    jd_int = int(np.floor(jd + 0.5))
    aa = jd_int + 32044
    bb = int(np.floor((4 * aa + 3) / 146097))
    cc = aa - int(np.floor((bb * 146097) / 4))
    dd = int(np.floor((4 * cc + 3) / 1461))
    ee = cc - int(np.floor((1461 * dd) / 4))
    mm = int(np.floor((5 * ee + 2) / 153))
    day = ee - int(np.floor((153 * mm + 2) / 5)) + 1
    month = mm + 3 - 12 * int(np.floor(mm / 10))
    year = bb * 100 + dd - 4800 + int(np.floor(mm / 10))
    # first check if the specified year is leap year or not (logical output)
    leapYear = (np.logical_or((np.mod(year,4) == np.logical_and(0,np.mod(year,100)) != 0),np.mod(year,400)) == 0)
    days = np.array([31,28,31,30,31,30,31,31,30,31,30,31])
    doy = sum(days[range(0,month-1)]) + day
    if leapYear == 1 and month > 2:
        doy = doy + 1
    
    doy = doy + mjd - int(np.floor(mjd))
    
    # determine the GPT3 coefficients
    
    # mean gravity in m/s**2
    gm = 9.80665
    # molar mass of dry air in kg/mol
    dMtr = 28.965 * 10 ** - 3
    # universal gas constant in J/K/mol
    Rg = 8.3143
    # factors for amplitudes
    if (it == 1):
        cosfy = 0
        coshy = 0
        sinfy = 0
        sinhy = 0
    else:
        cosfy = np.cos(doy / 365.25 * 2 * np.pi)
        coshy = np.cos(doy / 365.25 * 4 * np.pi)
        sinfy = np.sin(doy / 365.25 * 2 * np.pi)
        sinhy = np.sin(doy / 365.25 * 4 * np.pi)
    
    # read the respective data from the grid
    p_grid = grid[0]
    
    T_grid = grid[1]
    
    Q_grid = grid[2]
    
    dT_grid = grid[3]
    
    u_grid = grid[4]
    
    Hs_grid = grid[5]
    
    ah_grid = grid[6]
    
    aw_grid = grid[7]
    
    la_grid = grid[8]
    
    Tm_grid = grid[9]
    
    Gn_h_grid = grid[10]
    
    Ge_h_grid = grid[11]
    
    Gn_w_grid = grid[12]
    
    Ge_w_grid = grid[13]
    
    # determine the number of stations
    nstat = len(lat)
    # initialization
    p    = np.zeros(np.array([nstat,1]))
    T    = np.zeros(np.array([nstat,1]))
    dT   = np.zeros(np.array([nstat,1]))
    Tm   = np.zeros(np.array([nstat,1]))
    e    = np.zeros(np.array([nstat,1]))
    ah   = np.zeros(np.array([nstat,1]))
    aw   = np.zeros(np.array([nstat,1]))
    la   = np.zeros(np.array([nstat,1]))
    undu = np.zeros(np.array([nstat,1]))
    Gn_h = np.zeros(np.array([nstat,1]))
    Ge_h = np.zeros(np.array([nstat,1]))
    Gn_w = np.zeros(np.array([nstat,1]))
    Ge_w = np.zeros(np.array([nstat,1]))
    # loop over stations
    for k in range(0,nstat):
        # only positive longitude in degrees
        if lon[k] < 0:
            plon = (lon[k] + 2 * np.pi) * 180 / np.pi
        else:
            plon = lon[k] * 180 / np.pi
        # transform to polar distance in degrees
        ppod = (- lat[k] + np.pi / 2) * 180 / np.pi
        # find the index (line in the grid file) of the nearest point
        ipod = int(np.floor((ppod + 5) / 5))
        ilon = int(np.floor((plon + 5) / 5))
        # normalized (to one) differences, can be positive or negative
        diffpod = (ppod - (ipod * 5 - 2.5)) / 5
        difflon = (plon - (ilon * 5 - 2.5)) / 5
        if ipod == 36:
            ipod = 35
        if ilon == 72:
            ilon = 0
        if ilon == -1:
            ilon = 71
        indx=[]
        # get the number of the corresponding line
        indx.append((ipod - 1) * 72 + ilon-1)
        # near the poles: nearest neighbour interpolation, otherwise: bilinear
        bilinear = 0
        if ppod > 2.5 and ppod < 177.5:
            bilinear = 1
        # case of nearest neighbourhood
        if bilinear == 0:
            ix = indx[0]
            # transforming ellipsoidal height to orthometric height
            undu[k] = u_grid[ix]
            hgt = h_ell[k] - undu[k]
            # pressure, temperature at the height of the grid
            T0 = T_grid.iloc[ix,0] + T_grid.iloc[ix,1] * cosfy + T_grid.iloc[ix,2] * sinfy + T_grid.iloc[ix,3] * coshy + T_grid.iloc[ix,4] * sinhy
            p0 = p_grid.iloc[ix,0] + p_grid.iloc[ix,1] * cosfy + p_grid.iloc[ix,2] * sinfy + p_grid.iloc[ix,3] * coshy + p_grid.iloc[ix,4] * sinhy
            # specific humidity
            Q = Q_grid.iloc[ix,0] + Q_grid.iloc[ix,1] * cosfy + Q_grid.iloc[ix,2] * sinfy + Q_grid.iloc[ix,3] * coshy + Q_grid.iloc[ix,4] * sinhy
            # lapse rate of the temperature
            dT[k] = dT_grid.iloc[ix,0] + dT_grid.iloc[ix,1] * cosfy + dT_grid.iloc[ix,2] * sinfy + dT_grid.iloc[ix,3] * coshy + dT_grid.iloc[ix,4] * sinhy
            # station height - grid height
            redh = hgt - Hs_grid[ix]
            # temperature at station height in Celsius
            T[k] = T0 + dT[k] * redh - 273.15
            # temperature lapse rate in degrees / km
            dT[k] = dT[k] * 1000
            # virtual temperature in Kelvin
            Tv = T0 * (1 + 0.6077 * Q)
            c = gm * dMtr / (Rg * Tv)
            # pressure in hPa
            p[k] = (p0 * np.exp(- c * redh)) / 100
            # hydrostatic and wet coefficients ah and aw
            ah[k] = ah_grid.iloc[ix,0] + ah_grid.iloc[ix,1] * cosfy + ah_grid.iloc[ix,2] * sinfy + ah_grid.iloc[ix,3] * coshy + ah_grid.iloc[ix,4] * sinhy
            aw[k] = aw_grid.iloc[ix,0] + aw_grid.iloc[ix,1] * cosfy + aw_grid.iloc[ix,2] * sinfy + aw_grid.iloc[ix,3] * coshy + aw_grid.iloc[ix,4] * sinhy
            # water vapour decrease factor la
            la[k] = la_grid.iloc[ix,0] + la_grid.iloc[ix,1] * cosfy + la_grid.iloc[ix,2] * sinfy + la_grid.iloc[ix,3] * coshy + la_grid.iloc[ix,4] * sinhy
            # mean temperature Tm
            Tm[k] = Tm_grid.iloc[ix,0] + Tm_grid.iloc[ix,1] * cosfy + Tm_grid.iloc[ix,2] * sinfy + Tm_grid.iloc[ix,3] * coshy + Tm_grid.iloc[ix,4] * sinhy
            # north and east gradients (total, hydrostatic and wet)
            Gn_h[k] = Gn_h_grid.iloc[ix,0] + Gn_h_grid.iloc[ix,1] * cosfy + Gn_h_grid.iloc[ix,2] * sinfy + Gn_h_grid.iloc[ix,3] * coshy + Gn_h_grid.iloc[ix,4] * sinhy
            Ge_h[k] = Ge_h_grid.iloc[ix,0] + Ge_h_grid.iloc[ix,1] * cosfy + Ge_h_grid.iloc[ix,2] * sinfy + Ge_h_grid.iloc[ix,3] * coshy + Ge_h_grid.iloc[ix,4] * sinhy
            Gn_w[k] = Gn_w_grid.iloc[ix,0] + Gn_w_grid.iloc[ix,1] * cosfy + Gn_w_grid.iloc[ix,2] * sinfy + Gn_w_grid.iloc[ix,3] * coshy + Gn_w_grid.iloc[ix,4] * sinhy
            Ge_w[k] = Ge_w_grid.iloc[ix,0] + Ge_w_grid.iloc[ix,1] * cosfy + Ge_w_grid.iloc[ix,2] * sinfy + Ge_w_grid.iloc[ix,3] * coshy + Ge_w_grid.iloc[ix,4] * sinhy
            # water vapor pressure in hPa
            e0 = Q * p0 / (0.622 + 0.378 * Q) / 100
            e[k] = e0 * (100 * p[k] / p0) ** (la[k] + 1)
        else:
            ipod1 = int(ipod + np.sign(diffpod))
            ilon1 = int(ilon + np.sign(difflon))
            if ilon1 == 72:
                ilon1 = 0
            if ilon1 == -1:
                ilon1 = 71
            # get the number of the line
            indx.append((ipod1 - 1) * 72 + ilon-1)
            indx.append((ipod - 1) * 72 + ilon1-1)
            indx.append((ipod1 - 1) * 72 + ilon1-1)
            # transforming ellipsoidal height to orthometric height: Hortho = -N + Hell
            undul = u_grid[indx]
            hgt = h_ell[k] - undul
            # pressure, temperature at the height of the grid
            T0 = T_grid.iloc[indx,0] + T_grid.iloc[indx,1] * cosfy + T_grid.iloc[indx,2] * sinfy + T_grid.iloc[indx,3] * coshy + T_grid.iloc[indx,4] * sinhy
            p0 = p_grid.iloc[indx,0] + p_grid.iloc[indx,1] * cosfy + p_grid.iloc[indx,2] * sinfy + p_grid.iloc[indx,3] * coshy + p_grid.iloc[indx,4] * sinhy
            # humidity
            Ql = Q_grid.iloc[indx,0] + Q_grid.iloc[indx,1] * cosfy + Q_grid.iloc[indx,2] * sinfy + Q_grid.iloc[indx,3] * coshy + Q_grid.iloc[indx,4] * sinhy
            # reduction = stationheight - gridheight
            Hs1 = Hs_grid[indx]
            redh = hgt - Hs1
            # lapse rate of the temperature in degree / m
            dTl = dT_grid.iloc[indx,0] + dT_grid.iloc[indx,1] * cosfy + dT_grid.iloc[indx,2] * sinfy + dT_grid.iloc[indx,3] * coshy + dT_grid.iloc[indx,4] * sinhy
            # temperature reduction to station height
            Tl = T0 + np.multiply(dTl,redh) - 273.15
            # virtual temperature
            Tv = np.multiply(T0,(1 + 0.6077 * Ql))
            c = gm * dMtr / (Rg * Tv)
            # pressure in hPa
            pl = (np.multiply(p0,np.exp(np.multiply(- c,redh)))) / 100
            # hydrostatic and wet coefficients ah and aw
            ahl = ah_grid.iloc[indx,0] + ah_grid.iloc[indx,1] * cosfy + ah_grid.iloc[indx,2] * sinfy + ah_grid.iloc[indx,3] * coshy + ah_grid.iloc[indx,4] * sinhy
            awl = aw_grid.iloc[indx,0] + aw_grid.iloc[indx,1] * cosfy + aw_grid.iloc[indx,2] * sinfy + aw_grid.iloc[indx,3] * coshy + aw_grid.iloc[indx,4] * sinhy
            # water vapour decrease factor la
            lal = la_grid.iloc[indx,0] + la_grid.iloc[indx,1] * cosfy + la_grid.iloc[indx,2] * sinfy + la_grid.iloc[indx,3] * coshy + la_grid.iloc[indx,4] * sinhy
            # mean temperature of the water vapor Tm
            Tml = Tm_grid.iloc[indx,0] + Tm_grid.iloc[indx,1] * cosfy + Tm_grid.iloc[indx,2] * sinfy + Tm_grid.iloc[indx,3] * coshy + Tm_grid.iloc[indx,4] * sinhy
            # north and east gradients (total, hydrostatic and wet)
            Gn_hl = Gn_h_grid.iloc[indx,0] + Gn_h_grid.iloc[indx,1] * cosfy + Gn_h_grid.iloc[indx,2] * sinfy + Gn_h_grid.iloc[indx,3] * coshy + Gn_h_grid.iloc[indx,4] * sinhy
            Ge_hl = Ge_h_grid.iloc[indx,0] + Ge_h_grid.iloc[indx,1] * cosfy + Ge_h_grid.iloc[indx,2] * sinfy + Ge_h_grid.iloc[indx,3] * coshy + Ge_h_grid.iloc[indx,4] * sinhy
            Gn_wl = Gn_w_grid.iloc[indx,0] + Gn_w_grid.iloc[indx,1] * cosfy + Gn_w_grid.iloc[indx,2] * sinfy + Gn_w_grid.iloc[indx,3] * coshy + Gn_w_grid.iloc[indx,4] * sinhy
            Ge_wl = Ge_w_grid.iloc[indx,0] + Ge_w_grid.iloc[indx,1] * cosfy + Ge_w_grid.iloc[indx,2] * sinfy + Ge_w_grid.iloc[indx,3] * coshy + Ge_w_grid.iloc[indx,4] * sinhy
            # water vapor pressure in hPa
            e0 = np.multiply(Ql,p0) / (0.622 + 0.378 * Ql) / 100
            el = np.multiply(e0,(100.0 * pl / p0) ** (lal + 1))
            dnpod1 = np.abs(diffpod)
            dnpod2 = 1 - dnpod1
            dnlon1 = np.abs(difflon)
            dnlon2 = 1 - dnlon1
            # pressure
            R1 = dnpod2 * pl.iloc[0] + dnpod1 * pl.iloc[1]
            R2 = dnpod2 * pl.iloc[2] + dnpod1 * pl.iloc[3]
            p[k] = dnlon2 * R1 + dnlon1 * R2
            # temperature
            R1 = dnpod2 * Tl.iloc[0] + dnpod1 * Tl.iloc[1]
            R2 = dnpod2 * Tl.iloc[2] + dnpod1 * Tl.iloc[3]
            T[k] = dnlon2 * R1 + dnlon1 * R2
            # temperature in degree per km
            R1 = dnpod2 * dTl.iloc[0] + dnpod1 * dTl.iloc[1]
            R2 = dnpod2 * dTl.iloc[2] + dnpod1 * dTl.iloc[3]
            dT[k] = (dnlon2 * R1 + dnlon1 * R2) * 1000
            # water vapor pressure in hPa
            R1 = dnpod2 * el.iloc[0] + dnpod1 * el.iloc[1]
            R2 = dnpod2 * el.iloc[2] + dnpod1 * el.iloc[3]
            e[k] = dnlon2 * R1 + dnlon1 * R2
            # ah and aw
            R1 = dnpod2 * ahl.iloc[0] + dnpod1 * ahl.iloc[1]
            R2 = dnpod2 * ahl.iloc[2] + dnpod1 * ahl.iloc[3]
            ah[k] = dnlon2 * R1 + dnlon1 * R2
            R1 = dnpod2 * awl.iloc[0] + dnpod1 * awl.iloc[1]
            R2 = dnpod2 * awl.iloc[2] + dnpod1 * awl.iloc[3]
            aw[k] = dnlon2 * R1 + dnlon1 * R2
            # undulation
            R1 = dnpod2 * undul.iloc[0] + dnpod1 * undul.iloc[1]
            R2 = dnpod2 * undul.iloc[2] + dnpod1 * undul.iloc[3]
            undu[k] = dnlon2 * R1 + dnlon1 * R2
            # water vapor decrease factor la
            R1 = dnpod2 * lal.iloc[0] + dnpod1 * lal.iloc[1]
            R2 = dnpod2 * lal.iloc[2] + dnpod1 * lal.iloc[3]
            la[k] = dnlon2 * R1 + dnlon1 * R2
            # gradients
            R1 = dnpod2 * Gn_hl.iloc[0] + dnpod1 * Gn_hl.iloc[1]
            R2 = dnpod2 * Gn_hl.iloc[2] + dnpod1 * Gn_hl.iloc[3]
            Gn_h[k] = (dnlon2 * R1 + dnlon1 * R2)
            R1 = dnpod2 * Ge_hl.iloc[0] + dnpod1 * Ge_hl.iloc[1]
            R2 = dnpod2 * Ge_hl.iloc[2] + dnpod1 * Ge_hl.iloc[3]
            Ge_h[k] = (dnlon2 * R1 + dnlon1 * R2)
            R1 = dnpod2 * Gn_wl.iloc[0] + dnpod1 * Gn_wl.iloc[1]
            R2 = dnpod2 * Gn_wl.iloc[2] + dnpod1 * Gn_wl.iloc[3]
            Gn_w[k] = (dnlon2 * R1 + dnlon1 * R2)
            R1 = dnpod2 * Ge_wl.iloc[0] + dnpod1 * Ge_wl.iloc[1]
            R2 = dnpod2 * Ge_wl.iloc[2] + dnpod1 * Ge_wl.iloc[3]
            Ge_w[k] = (dnlon2 * R1 + dnlon1 * R2)
            # mean temperature of the water vapor Tm
            R1 = dnpod2 * Tml.iloc[0] + dnpod1 * Tml.iloc[1]
            R2 = dnpod2 * Tml.iloc[2] + dnpod1 * Tml.iloc[3]
            Tm[k] = dnlon2 * R1 + dnlon1 * R2
    
    return p,T,dT,Tm,e,ah,aw,la,undu,Gn_h,Ge_h,Gn_w,Ge_w




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Sep 10 20:09:27 2023

@author: psakicki
"""

#### Import star style
from geodezyx import *                   # Import the GeodeZYX modules
from geodezyx.externlib import *         # Import the external modules

P="/home/psakicki/GFZ_WORK/IPGP_WORK/REVOSIMA/0000_Pressure_Mayotte/010_from_Treden/RawData/transfer_3167556_files_8a99b7f8/204657_20210409_1130_data.txt"

DF = pd.read_table(P,sep=",")
#DF = DF.infer_objects()


#pd.to_datetime(DF.Time)


import pyrsktools

prsk = "/home/psakicki/GFZ_WORK/IPGP_WORK/REVOSIMA/0000_Pressure_Mayotte/011_RawData/010_A0A_RBR/204657_20210919_1458.rsk"
rsk = pyrsktools.RSK(prsk, readHiddenChannels=True)
rsk.open()



if 0:
    rsk._reader._query("data")
    
    t1 = np.datetime64("2021-03-03")
    t2 = np.datetime64("2021-06-04")
    
    rsk.readdata(t1,t2)


# DF = pd.DataFrame(rsk.data)


# outputDir="/home/psakicki/aaa_FOURBI/exportRSK"
# rsk.RSK2CSV(outputDir=outputDir)


# -*- coding: utf-8 -*-
"""
"""

import numpy as np
    
def vmf1_ht(ah = None,aw = None,dmjd = None,dlat = None,ht = None,zd = None): 
# =============================================================================
#     !!! This is the version with height correction !!!
#     !!! It has to be used with the grid !!!
#
#     This subroutine determines the VMF1 (Vienna Mapping Functions 1)
#     Reference: Boehm, J., B. Werl, H. Schuh (2006),
#     Troposphere mapping functions for GPS and very long baseline interferometry
#     from European Centre for Medium-Range Weather Forecasts operational analysis data,
#     J. Geoph. Res., Vol. 111, B02406, doi:10.1029/2005JB003629.
#
#     Please mind that the coefficients in this paper are wrong. The corrected version of
#     the paper can be found at:
#     http://ggosatm.hg.tuwien.ac.at/DOCS/PAPERS/2006Boehm_etal_VMF1.pdf
#
#     input data
#     ----------
#     ah:   hydrostatic coefficient a (http://ggosatm.hg.tuwien.ac.at/DELAY/GRID/)
#     aw:   wet coefficient a         (http://ggosatm.hg.tuwien.ac.at/DELAY/GRID/)
#     dmjd: modified julian date
#     dlat: ellipsoidal latitude in radians
#     ht:   ellipsoidal height in meter
#     zd:   zenith distance in radians
#
#     output data
#     -----------
#     vmf1h: hydrostatic mapping function
#     vmf1w: wet mapping function
#
#     Johannes Boehm, 2005 October 2
#      Rev 2011 July 21: latitude -> ellipsoidal latitude

#     implicit double precision (a-h,o-z)
    
#     pi = 3.14159265359d0
    
#     reference day is 28 January
#     this is taken from Niell (1996) to be consistent
#     File Converted to python script by Zohreh Adavi zohreh.adavi@tuwien.ac.at, 2023-02-20
# =============================================================================  
    doy = dmjd - 44239.0 + 1 - 28
    bh = 0.0029
    c0h = 0.062
    if (dlat < 0):
        phh = np.pi
        c11h = 0.007
        c10h = 0.002
    else:
        phh = 0.0
        c11h = 0.005
        c10h = 0.001
    
    ch = c0h + ((np.cos(doy / 365.25 * 2.0 * np.pi + phh) + 1.0) * c11h / 2.0 + c10h) * (1.0 - np.cos(dlat))
    sine = np.sin(np.pi / 2.0 - zd)
    beta = bh / (sine + ch)
    gamma = ah / (sine + beta)
    topcon = (1.0 + ah / (1.0 + bh / (1.0 + ch)))
    vmf1h = topcon / (sine + gamma)
    # C  height correction for hydrotatic part [Niell, 1996]
    a_ht = 2.53e-05
    b_ht = 0.00549
    c_ht = 0.00114
    hs_km = ht / 1000.0
    beta = b_ht / (sine + c_ht)
    gamma = a_ht / (sine + beta)
    topcon = (1.0 + a_ht / (1.0 + b_ht / (1.0 + c_ht)))
    ht_corr_coef = 1.0 / sine - topcon / (sine + gamma)
    ht_corr = np.multiply(ht_corr_coef,hs_km)
    vmf1h = vmf1h + ht_corr
    bw = 0.00146
    cw = 0.04391
    beta = bw / (sine + cw)
    gamma = aw / (sine + beta)
    topcon = (1.0 + aw / (1.0 + bw / (1.0 + cw)))
    vmf1w = topcon / (sine + gamma)
    return vmf1h,vmf1w